"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""


from keras_core.ops import image
from keras_core.ops import nn
from keras_core.ops import numpy
from keras_core.src.ops.core import cast
from keras_core.src.ops.core import cond
from keras_core.src.ops.core import convert_to_numpy
from keras_core.src.ops.core import convert_to_tensor
from keras_core.src.ops.core import fori_loop
from keras_core.src.ops.core import scatter
from keras_core.src.ops.core import scatter_update
from keras_core.src.ops.core import shape
from keras_core.src.ops.core import slice
from keras_core.src.ops.core import slice_update
from keras_core.src.ops.core import stop_gradient
from keras_core.src.ops.core import unstack
from keras_core.src.ops.core import while_loop
from keras_core.src.ops.math import fft
from keras_core.src.ops.math import fft2
from keras_core.src.ops.math import in_top_k
from keras_core.src.ops.math import logsumexp
from keras_core.src.ops.math import qr
from keras_core.src.ops.math import segment_sum
from keras_core.src.ops.math import top_k
from keras_core.src.ops.nn import average_pool
from keras_core.src.ops.nn import binary_crossentropy
from keras_core.src.ops.nn import categorical_crossentropy
from keras_core.src.ops.nn import conv
from keras_core.src.ops.nn import conv_transpose
from keras_core.src.ops.nn import depthwise_conv
from keras_core.src.ops.nn import elu
from keras_core.src.ops.nn import gelu
from keras_core.src.ops.nn import hard_sigmoid
from keras_core.src.ops.nn import leaky_relu
from keras_core.src.ops.nn import log_sigmoid
from keras_core.src.ops.nn import log_softmax
from keras_core.src.ops.nn import max_pool
from keras_core.src.ops.nn import multi_hot
from keras_core.src.ops.nn import one_hot
from keras_core.src.ops.nn import relu
from keras_core.src.ops.nn import relu6
from keras_core.src.ops.nn import selu
from keras_core.src.ops.nn import separable_conv
from keras_core.src.ops.nn import sigmoid
from keras_core.src.ops.nn import silu
from keras_core.src.ops.nn import softmax
from keras_core.src.ops.nn import softplus
from keras_core.src.ops.nn import softsign
from keras_core.src.ops.nn import sparse_categorical_crossentropy
from keras_core.src.ops.nn import swish
from keras_core.src.ops.nn import tanh
from keras_core.src.ops.numpy import abs
from keras_core.src.ops.numpy import absolute
from keras_core.src.ops.numpy import add
from keras_core.src.ops.numpy import all
from keras_core.src.ops.numpy import amax
from keras_core.src.ops.numpy import amin
from keras_core.src.ops.numpy import any
from keras_core.src.ops.numpy import append
from keras_core.src.ops.numpy import arange
from keras_core.src.ops.numpy import arccos
from keras_core.src.ops.numpy import arcsin
from keras_core.src.ops.numpy import arctan
from keras_core.src.ops.numpy import arctan2
from keras_core.src.ops.numpy import argmax
from keras_core.src.ops.numpy import argmin
from keras_core.src.ops.numpy import argsort
from keras_core.src.ops.numpy import array
from keras_core.src.ops.numpy import average
from keras_core.src.ops.numpy import bincount
from keras_core.src.ops.numpy import broadcast_to
from keras_core.src.ops.numpy import ceil
from keras_core.src.ops.numpy import clip
from keras_core.src.ops.numpy import concatenate
from keras_core.src.ops.numpy import conj
from keras_core.src.ops.numpy import conjugate
from keras_core.src.ops.numpy import copy
from keras_core.src.ops.numpy import cos
from keras_core.src.ops.numpy import count_nonzero
from keras_core.src.ops.numpy import cross
from keras_core.src.ops.numpy import cumprod
from keras_core.src.ops.numpy import cumsum
from keras_core.src.ops.numpy import diag
from keras_core.src.ops.numpy import diagonal
from keras_core.src.ops.numpy import divide
from keras_core.src.ops.numpy import dot
from keras_core.src.ops.numpy import einsum
from keras_core.src.ops.numpy import empty
from keras_core.src.ops.numpy import equal
from keras_core.src.ops.numpy import exp
from keras_core.src.ops.numpy import expand_dims
from keras_core.src.ops.numpy import expm1
from keras_core.src.ops.numpy import eye
from keras_core.src.ops.numpy import flip
from keras_core.src.ops.numpy import floor
from keras_core.src.ops.numpy import floor_divide
from keras_core.src.ops.numpy import full
from keras_core.src.ops.numpy import full_like
from keras_core.src.ops.numpy import get_item
from keras_core.src.ops.numpy import greater
from keras_core.src.ops.numpy import greater_equal
from keras_core.src.ops.numpy import hstack
from keras_core.src.ops.numpy import identity
from keras_core.src.ops.numpy import imag
from keras_core.src.ops.numpy import isclose
from keras_core.src.ops.numpy import isfinite
from keras_core.src.ops.numpy import isinf
from keras_core.src.ops.numpy import isnan
from keras_core.src.ops.numpy import less
from keras_core.src.ops.numpy import less_equal
from keras_core.src.ops.numpy import linspace
from keras_core.src.ops.numpy import log
from keras_core.src.ops.numpy import log10
from keras_core.src.ops.numpy import log1p
from keras_core.src.ops.numpy import log2
from keras_core.src.ops.numpy import logaddexp
from keras_core.src.ops.numpy import logical_and
from keras_core.src.ops.numpy import logical_not
from keras_core.src.ops.numpy import logical_or
from keras_core.src.ops.numpy import logical_xor
from keras_core.src.ops.numpy import logspace
from keras_core.src.ops.numpy import matmul
from keras_core.src.ops.numpy import max
from keras_core.src.ops.numpy import maximum
from keras_core.src.ops.numpy import mean
from keras_core.src.ops.numpy import meshgrid
from keras_core.src.ops.numpy import min
from keras_core.src.ops.numpy import minimum
from keras_core.src.ops.numpy import mod
from keras_core.src.ops.numpy import moveaxis
from keras_core.src.ops.numpy import multiply
from keras_core.src.ops.numpy import nan_to_num
from keras_core.src.ops.numpy import ndim
from keras_core.src.ops.numpy import negative
from keras_core.src.ops.numpy import nonzero
from keras_core.src.ops.numpy import not_equal
from keras_core.src.ops.numpy import ones
from keras_core.src.ops.numpy import ones_like
from keras_core.src.ops.numpy import outer
from keras_core.src.ops.numpy import pad
from keras_core.src.ops.numpy import power
from keras_core.src.ops.numpy import prod
from keras_core.src.ops.numpy import ravel
from keras_core.src.ops.numpy import real
from keras_core.src.ops.numpy import reciprocal
from keras_core.src.ops.numpy import repeat
from keras_core.src.ops.numpy import reshape
from keras_core.src.ops.numpy import roll
from keras_core.src.ops.numpy import round
from keras_core.src.ops.numpy import sign
from keras_core.src.ops.numpy import sin
from keras_core.src.ops.numpy import size
from keras_core.src.ops.numpy import sort
from keras_core.src.ops.numpy import split
from keras_core.src.ops.numpy import sqrt
from keras_core.src.ops.numpy import square
from keras_core.src.ops.numpy import squeeze
from keras_core.src.ops.numpy import stack
from keras_core.src.ops.numpy import std
from keras_core.src.ops.numpy import subtract
from keras_core.src.ops.numpy import sum
from keras_core.src.ops.numpy import swapaxes
from keras_core.src.ops.numpy import take
from keras_core.src.ops.numpy import take_along_axis
from keras_core.src.ops.numpy import tan
from keras_core.src.ops.numpy import tensordot
from keras_core.src.ops.numpy import tile
from keras_core.src.ops.numpy import trace
from keras_core.src.ops.numpy import transpose
from keras_core.src.ops.numpy import tri
from keras_core.src.ops.numpy import tril
from keras_core.src.ops.numpy import triu
from keras_core.src.ops.numpy import true_divide
from keras_core.src.ops.numpy import var
from keras_core.src.ops.numpy import vdot
from keras_core.src.ops.numpy import vstack
from keras_core.src.ops.numpy import where
from keras_core.src.ops.numpy import zeros
from keras_core.src.ops.numpy import zeros_like



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""


from keras_core.optimizers import schedules
from keras_core.src.optimizers import deserialize
from keras_core.src.optimizers import get
from keras_core.src.optimizers import serialize
from keras_core.src.optimizers.adadelta import Adadelta
from keras_core.src.optimizers.adafactor import Adafactor
from keras_core.src.optimizers.adagrad import Adagrad
from keras_core.src.optimizers.adam import Adam
from keras_core.src.optimizers.adamax import Adamax
from keras_core.src.optimizers.adamw import AdamW
from keras_core.src.optimizers.ftrl import Ftrl
from keras_core.src.optimizers.nadam import Nadam
from keras_core.src.optimizers.optimizer import Optimizer
from keras_core.src.optimizers.rmsprop import RMSprop
from keras_core.src.optimizers.sgd import SGD



import jax
import numpy as np
import tree

from keras_core.src import backend
from keras_core.src import callbacks as callbacks_module
from keras_core.src import ops
from keras_core.src import optimizers as optimizers_module
from keras_core.src.trainers import trainer as base_trainer
from keras_core.src.trainers.data_adapters import data_adapter_utils
from keras_core.src.trainers.epoch_iterator import EpochIterator
from keras_core.src.utils import traceback_utils


class JAXTrainer(base_trainer.Trainer):
    def __init__(self):
        super().__init__()
        self.train_function = None
        self.test_function = None
        self.predict_function = None

    def compute_loss_and_updates(
        self,
        trainable_variables,
        non_trainable_variables,
        x,
        y,
        sample_weight,
        training=False,
    ):
        """This method is stateless and is intended for use with jax.grad."""
        kwargs = {}
        if self._call_has_training_arg:
            kwargs["training"] = training
        y_pred, non_trainable_variables, losses = self.stateless_call(
            trainable_variables,
            non_trainable_variables,
            x,
            return_losses=True,
            **kwargs,
        )
        loss = self.compute_loss(x, y, y_pred, sample_weight, allow_empty=True)
        if losses:
            loss += ops.sum(losses)
        return loss, (y_pred, non_trainable_variables)

    def _eager_build(self, data_batch):
        model_unbuilt = not all(layer.built for layer in self._flatten_layers())
        compile_metrics_unbuilt = (
            self._compile_metrics is not None
            and not self._compile_metrics.built
        )
        if model_unbuilt or compile_metrics_unbuilt:
            # Build the model on one batch of data.
            (
                x,
                y,
                sample_weight,
            ) = data_adapter_utils.unpack_x_y_sample_weight(data_batch)
            # Build model
            with backend.StatelessScope():
                y_pred = self(x)
                if compile_metrics_unbuilt:
                    # Build metrics
                    self.compute_metrics(
                        x, y, y_pred, sample_weight=sample_weight
                    )
        if self.optimizer is not None and not self.optimizer.built:
            # Build optimizer
            self.optimizer.build(self.trainable_variables)

    def train_step(self, state, data):
        (
            trainable_variables,
            non_trainable_variables,
            optimizer_variables,
            metrics_variables,
        ) = state
        x, y, sample_weight = data_adapter_utils.unpack_x_y_sample_weight(data)
        grad_fn = jax.value_and_grad(
            self.compute_loss_and_updates, has_aux=True
        )
        (loss, (y_pred, non_trainable_variables)), grads = grad_fn(
            trainable_variables,
            non_trainable_variables,
            x,
            y,
            sample_weight,
            training=True,
        )

        (
            trainable_variables,
            optimizer_variables,
        ) = self.optimizer.stateless_apply(
            optimizer_variables, grads, trainable_variables
        )

        with backend.StatelessScope(
            state_mapping=[
                (ref_v, v)
                for ref_v, v in zip(self.metrics_variables, metrics_variables)
            ]
        ) as scope:
            self._loss_tracker.update_state(loss)
            logs = self.compute_metrics(x, y, y_pred, sample_weight)

        new_metrics_variables = []
        for ref_v in self.metrics_variables:
            new_v = scope.get_current_value(ref_v)
            if new_v is None:
                new_v = ref_v.value
            new_metrics_variables.append(new_v)
        metrics_variables = new_metrics_variables

        state = (
            trainable_variables,
            non_trainable_variables,
            optimizer_variables,
            metrics_variables,
        )
        return logs, state

    def test_step(self, state, data):
        (
            trainable_variables,
            non_trainable_variables,
            metrics_variables,
        ) = state
        x, y, sample_weight = data_adapter_utils.unpack_x_y_sample_weight(data)
        loss, (
            y_pred,
            non_trainable_variables,
        ) = self.compute_loss_and_updates(
            trainable_variables,
            non_trainable_variables,
            x,
            y,
            sample_weight,
            training=False,
        )

        with backend.StatelessScope(
            state_mapping=[
                (ref_v, v)
                for ref_v, v in zip(self.metrics_variables, metrics_variables)
            ]
        ) as scope:
            self._loss_tracker.update_state(loss)
            logs = self.compute_metrics(x, y, y_pred, sample_weight)

        new_metrics_variables = []
        for ref_v in self.metrics_variables:
            new_v = scope.get_current_value(ref_v)
            if new_v is None:
                new_v = ref_v.value
            new_metrics_variables.append(new_v)
        metrics_variables = new_metrics_variables

        state = (
            trainable_variables,
            non_trainable_variables,
            metrics_variables,
        )
        return logs, state

    def predict_step(self, state, data):
        trainable_variables, non_trainable_variables = state
        kwargs = {}
        if self._call_has_training_arg:
            kwargs["training"] = False

        x, _, _ = data_adapter_utils.unpack_x_y_sample_weight(data)
        outputs, non_trainable_variables = self.stateless_call(
            trainable_variables, non_trainable_variables, x, **kwargs
        )
        return outputs, (trainable_variables, non_trainable_variables)

    def make_train_function(self, force=False):
        if self.train_function is not None and not force:
            return self.train_function

        def one_train_step(state, data):
            data = data[0]
            return self.train_step(state, data)

        def multi_train_steps(state, data):
            for single_step_data in data:
                logs, state = one_train_step(state, [single_step_data])
            return logs, state

        if self.steps_per_execution > 1:
            train_step = multi_train_steps
        else:
            train_step = one_train_step

        if not self.run_eagerly and self.jit_compile:

            @jax.jit
            def compiled_train_step(state, data):
                return train_step(state, data)

            self.train_function = compiled_train_step

        else:
            self.train_function = train_step

    def make_test_function(self, force=False):
        if self.test_function is not None and not force:
            return self.test_function

        def one_test_step(state, data):
            data = data[0]
            return self.test_step(state, data)

        def multi_test_steps(state, data):
            for single_step_data in data:
                logs, state = one_test_step(state, [single_step_data])
            return logs, state

        if self.steps_per_execution > 1:
            test_step = multi_test_steps
        else:
            test_step = one_test_step

        if not self.run_eagerly and self.jit_compile:

            @jax.jit
            def compiled_test_step(state, data):
                return test_step(state, data)

            self.test_function = compiled_test_step

        else:
            self.test_function = test_step

    def make_predict_function(self, force=False):
        if self.predict_function is not None and not force:
            return self.predict_function

        def one_predict_step(state, data):
            data = data[0]
            return self.predict_step(state, data)

        def multi_predict_steps(state, data):
            outputs, state = one_predict_step(state, data[:1])

            for single_step_data in data[1:]:
                step_outputs, state = one_predict_step(
                    state,
                    [single_step_data],
                )
                outputs = tree.map_structure(
                    lambda t1, t2: jax.numpy.concatenate([t1, t2]),
                    outputs,
                    step_outputs,
                )
            return outputs, state

        if self.steps_per_execution > 1:
            predict_step = multi_predict_steps
        else:
            predict_step = one_predict_step

        if not self.run_eagerly and self.jit_compile:

            @jax.jit
            def compiled_predict_step(state, data):
                return predict_step(state, data)

            self.predict_function = compiled_predict_step

        else:
            self.predict_function = predict_step

    @traceback_utils.filter_traceback
    def fit(
        self,
        x=None,
        y=None,
        batch_size=None,
        epochs=1,
        verbose="auto",
        callbacks=None,
        validation_split=0.0,
        validation_data=None,
        shuffle=True,
        class_weight=None,
        sample_weight=None,
        initial_epoch=0,
        steps_per_epoch=None,
        validation_steps=None,
        validation_batch_size=None,
        validation_freq=1,
    ):
        self._assert_compile_called("fit")
        # TODO: respect compiled trainable state
        if validation_split and validation_data is None:
            # Create the validation data using the training data. Only supported
            # for TF/numpy/jax arrays.
            (
                x,
                y,
                sample_weight,
            ), validation_data = data_adapter_utils.train_validation_split(
                (x, y, sample_weight), validation_split=validation_split
            )

        if validation_data:
            (
                val_x,
                val_y,
                val_sample_weight,
            ) = data_adapter_utils.unpack_x_y_sample_weight(validation_data)

        # Create an iterator that yields batches for one epoch.
        epoch_iterator = EpochIterator(
            x=x,
            y=y,
            sample_weight=sample_weight,
            batch_size=batch_size,
            steps_per_epoch=steps_per_epoch,
            shuffle=shuffle,
            class_weight=class_weight,
            steps_per_execution=self.steps_per_execution,
        )

        needs_building = (
            not all(layer.built for layer in self._flatten_layers())
            or not self.optimizer.built
            or (
                self._compile_metrics is not None
                and not self._compile_metrics.built
            )
        )
        if needs_building:
            # Build the model on one batch of data.
            for _, data in epoch_iterator.enumerate_epoch(return_type="np"):
                data_batch = data[0]
                self._eager_build(data_batch)
                break

        # Container that configures and calls callbacks.
        if not isinstance(callbacks, callbacks_module.CallbackList):
            callbacks = callbacks_module.CallbackList(
                callbacks,
                add_history=True,
                add_progbar=verbose != 0,
                verbose=verbose,
                epochs=epochs,
                steps=epoch_iterator.num_batches,
                model=self,
            )

        self.make_train_function()
        self.stop_training = False
        callbacks.on_train_begin()

        for epoch in range(initial_epoch, epochs):
            self.reset_metrics()
            callbacks.on_epoch_begin(epoch)

            trainable_variables = self.trainable_variables
            non_trainable_variables = self.non_trainable_variables
            optimizer_variables = self.optimizer.variables
            metrics_variables = self.metrics_variables

            for step, data in epoch_iterator.enumerate_epoch(return_type="np"):
                # Callbacks
                callbacks.on_train_batch_begin(step)

                # Train step
                state = (
                    trainable_variables,
                    non_trainable_variables,
                    optimizer_variables,
                    metrics_variables,
                )
                logs, state = self.train_function(state, data)
                (
                    trainable_variables,
                    non_trainable_variables,
                    optimizer_variables,
                    metrics_variables,
                ) = state

                # Setting _jax_state enables callbacks to force a state sync
                # if they need to.
                self._jax_state = {
                    "trainable_variables": trainable_variables,
                    "non_trainable_variables": non_trainable_variables,
                    "optimizer_variables": optimizer_variables,
                    "metrics_variables": metrics_variables,
                }

                # Callbacks
                callbacks.on_train_batch_end(step, self._pythonify_logs(logs))
                if self.stop_training:
                    break

            # Reattach state to model variables.
            # NOTE: doing this after each step would be a big performance
            # bottleneck.
            self.jax_state_sync()

            # Override with model metrics instead of last step logs
            epoch_logs = self.get_metrics_result()

            # Run validation.
            if validation_data and self._should_eval(epoch, validation_freq):
                # Create EpochIterator for evaluation and cache it.
                if getattr(self, "_eval_epoch_iterator", None) is None:
                    self._eval_epoch_iterator = EpochIterator(
                        x=val_x,
                        y=val_y,
                        sample_weight=val_sample_weight,
                        batch_size=validation_batch_size or batch_size,
                        steps_per_execution=self.steps_per_execution,
                    )
                val_logs = self.evaluate(
                    x=val_x,
                    y=val_y,
                    sample_weight=val_sample_weight,
                    batch_size=validation_batch_size or batch_size,
                    steps=validation_steps,
                    callbacks=callbacks,
                    return_dict=True,
                    _use_cached_eval_dataset=True,
                )
                val_logs = {
                    "val_" + name: val for name, val in val_logs.items()
                }
                epoch_logs.update(val_logs)

            callbacks.on_epoch_end(epoch, epoch_logs)
            training_logs = epoch_logs
            if self.stop_training:
                break

        if (
            isinstance(self.optimizer, optimizers_module.Optimizer)
            and epochs > 0
        ):
            self.optimizer.finalize_variable_values(self.trainable_weights)

        # If _eval_epoch_iterator exists, delete it after all epochs are done.
        if getattr(self, "_eval_epoch_iterator", None) is not None:
            del self._eval_epoch_iterator
        callbacks.on_train_end(logs=training_logs)
        self._jax_state = None
        return self.history

    @traceback_utils.filter_traceback
    def evaluate(
        self,
        x=None,
        y=None,
        batch_size=None,
        verbose="auto",
        sample_weight=None,
        steps=None,
        callbacks=None,
        return_dict=False,
        **kwargs,
    ):
        self._assert_compile_called("evaluate")
        # TODO: respect compiled trainable state
        use_cached_eval_dataset = kwargs.pop("_use_cached_eval_dataset", False)
        if kwargs:
            raise ValueError(f"Arguments not recognized: {kwargs}")

        if use_cached_eval_dataset:
            epoch_iterator = self._eval_epoch_iterator
        else:
            # Create an iterator that yields batches of input/target data.
            epoch_iterator = EpochIterator(
                x=x,
                y=y,
                sample_weight=sample_weight,
                batch_size=batch_size,
                steps_per_epoch=steps,
                shuffle=False,
                steps_per_execution=self.steps_per_execution,
            )

        if not all(layer.built for layer in self._flatten_layers()):
            # Build the model on one batch of data.
            for _, data in epoch_iterator.enumerate_epoch(return_type="np"):
                data_batch = data[0]
                self._eager_build(data_batch)
                break

        # Container that configures and calls callbacks.
        if not isinstance(callbacks, callbacks_module.CallbackList):
            callbacks = callbacks_module.CallbackList(
                callbacks,
                add_history=True,
                add_progbar=verbose != 0,
                verbose=verbose,
                epochs=1,
                steps=epoch_iterator.num_batches,
                model=self,
            )

        self.make_test_function()
        callbacks.on_test_begin()
        logs = None
        self.reset_metrics()

        trainable_variables = self.trainable_variables
        non_trainable_variables = self.non_trainable_variables
        metrics_variables = self.metrics_variables

        for step, data in epoch_iterator.enumerate_epoch(return_type="np"):
            callbacks.on_test_batch_begin(step)

            state = (
                trainable_variables,
                non_trainable_variables,
                metrics_variables,
            )
            logs, state = self.test_function(state, data)
            # Note that trainable variables are not returned since they're
            # immutable here.
            _, non_trainable_variables, metrics_variables = state

            # Setting _jax_state enables callbacks to force a state sync
            # if they need to.
            self._jax_state = {
                # I wouldn't recommend modifying non-trainable model state
                # during evaluate(), but it's allowed.
                "non_trainable_variables": non_trainable_variables,
                "metrics_variables": metrics_variables,
            }
            callbacks.on_test_batch_end(step, self._pythonify_logs(logs))

        # Reattach state back to model.
        self.jax_state_sync()

        logs = self.get_metrics_result()
        callbacks.on_test_end(logs)
        self._jax_state = None
        if return_dict:
            return logs
        return self._flatten_metrics_in_order(logs)

    @traceback_utils.filter_traceback
    def predict(
        self, x, batch_size=None, verbose="auto", steps=None, callbacks=None
    ):
        # Create an iterator that yields batches of input data.
        epoch_iterator = EpochIterator(
            x=x,
            batch_size=batch_size,
            steps_per_epoch=steps,
            shuffle=False,
            steps_per_execution=self.steps_per_execution,
        )

        if not all(layer.built for layer in self._flatten_layers()):
            # Build the model on one batch of data.
            for _, data in epoch_iterator.enumerate_epoch(return_type="np"):
                # Build model
                x, _, _ = data_adapter_utils.unpack_x_y_sample_weight(data[0])
                with backend.StatelessScope():
                    self(x)
                break

        # Container that configures and calls callbacks.
        if not isinstance(callbacks, callbacks_module.CallbackList):
            callbacks = callbacks_module.CallbackList(
                callbacks,
                add_history=True,
                add_progbar=verbose != 0,
                verbose=verbose,
                epochs=1,
                steps=epoch_iterator.num_batches,
                model=self,
            )

        self.make_predict_function()
        callbacks.on_predict_begin()

        def append_to_outputs(batch_outputs, outputs):
            if outputs is None:
                outputs = tree.map_structure(
                    lambda batch_output: [batch_output],
                    batch_outputs,
                )
            else:
                tree.map_structure_up_to(
                    batch_outputs,
                    lambda output, batch_output: output.append(batch_output),
                    outputs,
                    batch_outputs,
                )
            return outputs

        trainable_variables = self.trainable_variables
        non_trainable_variables = self.non_trainable_variables
        state = (trainable_variables, non_trainable_variables)
        outputs = None
        for step, x in epoch_iterator.enumerate_epoch(return_type="np"):
            callbacks.on_predict_batch_begin(step)
            batch_outputs, state = self.predict_function(state, x)
            outputs = append_to_outputs(batch_outputs, outputs)
            callbacks.on_predict_batch_end(step, {"outputs": batch_outputs})
        callbacks.on_predict_end()
        return tree.map_structure_up_to(batch_outputs, np.concatenate, outputs)

    def train_on_batch(
        self,
        x,
        y=None,
        sample_weight=None,
        class_weight=None,
        return_dict=False,
    ):
        self._assert_compile_called("train_on_batch")
        if class_weight is not None:
            if sample_weight is not None:
                raise ValueError(
                    "Arguments `sample_weight` and `class_weight` "
                    "cannot be specified at the same time. "
                    f"Received: sample_weight={sample_weight}, "
                    f"class_weight={class_weight}"
                )
            sample_weight = data_adapter_utils.class_weight_to_sample_weights(
                y, class_weight
            )
        data = (x, y, sample_weight)

        # Maybe build model
        self._eager_build(data)
        self.make_train_function()

        # Train step
        trainable_variables = self.trainable_variables
        non_trainable_variables = self.non_trainable_variables
        optimizer_variables = self.optimizer.variables
        metrics_variables = self.metrics_variables
        state = (
            trainable_variables,
            non_trainable_variables,
            optimizer_variables,
            metrics_variables,
        )
        logs, state = self.train_function(state, [data])

        # State sync
        (
            trainable_variables,
            non_trainable_variables,
            optimizer_variables,
            metrics_variables,
        ) = state
        self._jax_state = {
            "trainable_variables": trainable_variables,
            "non_trainable_variables": non_trainable_variables,
            "optimizer_variables": optimizer_variables,
            "metrics_variables": metrics_variables,
        }
        self.jax_state_sync()

        # Format return values
        logs = tree.map_structure(lambda x: np.array(x), logs)
        if return_dict:
            return logs
        return self._flatten_metrics_in_order(logs)

    def test_on_batch(
        self,
        x,
        y=None,
        sample_weight=None,
        return_dict=False,
    ):
        self._assert_compile_called("test_on_batch")

        data = (x, y, sample_weight)
        # Maybe build model
        self._eager_build(data)
        self.make_test_function()

        # Test step
        trainable_variables = self.trainable_variables
        non_trainable_variables = self.non_trainable_variables
        metrics_variables = self.metrics_variables
        state = (
            trainable_variables,
            non_trainable_variables,
            metrics_variables,
        )
        logs, state = self.test_function(state, [data])

        # State sync
        _, non_trainable_variables, metrics_variables = state
        self._jax_state = {
            "non_trainable_variables": non_trainable_variables,
            "metrics_variables": metrics_variables,
        }
        self.jax_state_sync()

        # Format return values.
        logs = tree.map_structure(lambda x: np.array(x), logs)
        if return_dict:
            return logs
        return self._flatten_metrics_in_order(logs)

    def predict_on_batch(self, x):
        if not all(layer.built for layer in self._flatten_layers()):
            # Build model
            with backend.StatelessScope():
                self(x)

        self.make_predict_function()
        trainable_variables = self.trainable_variables
        non_trainable_variables = self.non_trainable_variables
        state = (trainable_variables, non_trainable_variables)
        batch_outputs, state = self.predict_function(state, [x])
        batch_outputs = tree.map_structure(lambda x: np.array(x), batch_outputs)
        return batch_outputs

    def jax_state_sync(self):
        if not getattr(self, "_jax_state", None):
            return

        trainable_variables = self._jax_state.get("trainable_variables", None)
        non_trainable_variables = self._jax_state.get(
            "non_trainable_variables", None
        )
        optimizer_variables = self._jax_state.get("optimizer_variables", None)
        metrics_variables = self._jax_state.get("metrics_variables", None)
        if trainable_variables:
            for ref_v, v in zip(self.trainable_variables, trainable_variables):
                ref_v.assign(v)
        if non_trainable_variables:
            for ref_v, v in zip(
                self.non_trainable_variables, non_trainable_variables
            ):
                ref_v.assign(v)
        if optimizer_variables:
            for ref_v, v in zip(self.optimizer.variables, optimizer_variables):
                ref_v.assign(v)
        if metrics_variables:
            for ref_v, v in zip(self.metrics_variables, metrics_variables):
                ref_v.assign(v)




import random as python_random

import numpy as np

from keras_core.src import backend
from keras_core.src.api_export import keras_core_export


@keras_core_export("keras_core.random.SeedGenerator")
class SeedGenerator:
    """Generates variable seeds upon each call to a RNG-using function.

    In Keras, all RNG-using methods (such as `keras_core.random.normal()`)
    are stateless, meaning that if you pass an integer seed to them
    (such as `seed=42`), they will return the same values at each call.
    In order to get different values at each call, you must use a
    `SeedGenerator` instead as the seed argument. The `SeedGenerator`
    instead is stateful.

    Example:

    ```python
    seed_gen = keras_core.random.SeedGenerator(seed=42)
    values = keras_core.random.normal(shape=(2, 3), seed=seed_gen)
    new_values = keras_core.random.normal(shape=(2, 3), seed=seed_gen)
    ```
    """

    def __init__(self, seed, **kwargs):
        custom_backend = kwargs.pop("backend", None)
        if kwargs:
            raise ValueError(f"Unrecognized keyword arguments: {kwargs}")
        if custom_backend is not None:
            self.backend = custom_backend
        else:
            self.backend = backend

        self._initial_seed = seed
        if seed is None:
            seed = make_default_seed()

        if not isinstance(seed, int):
            raise ValueError(
                "Argument `seed` must be an integer. " f"Received: seed={seed}"
            )

        def seed_initializer(*args, **kwargs):
            dtype = kwargs.get("dtype", None)
            return self.backend.convert_to_tensor([seed, 0], dtype=dtype)

        self.state = self.backend.Variable(
            seed_initializer,
            shape=(2,),
            dtype="uint32",
            trainable=False,
            name="seed_generator_state",
        )


def make_default_seed():
    return python_random.randint(1, int(1e9))


def draw_seed(seed):
    from keras_core.src.backend import convert_to_tensor

    if isinstance(seed, SeedGenerator):
        seed_state = seed.state
        # Use * 1 to create a copy
        new_seed_value = seed_state.value * 1
        increment = seed.backend.convert_to_tensor(
            np.array([0, 1]), dtype="uint32"
        )
        seed.state.assign(seed_state + increment)
        return new_seed_value
    elif isinstance(seed, int):
        return convert_to_tensor([seed, 0], dtype="uint32")
    elif seed is None:
        return convert_to_tensor([make_default_seed(), 0], dtype="uint32")
    raise ValueError(
        "Argument `seed` must be either an integer "
        "or an instance of `SeedGenerator`. "
        f"Received: seed={seed} (of type {type(seed)})"
    )




from keras_core.src import backend
from keras_core.src import ops


def dtype_size(dtype):
    if dtype in ("bfloat16", "float16"):
        return 16
    if dtype in ("float32", "int32"):
        return 32
    if dtype in ("float64", "int64"):
        return 64
    if dtype == "uint8":
        return 8
    if dtype == "bool":
        return 1
    raise ValueError(f"Invalid dtype: {dtype}")


def is_float(dtype):
    return "float" in dtype


def cast_to_common_dtype(tensors):
    """Cast a list of tensors to a common dtype.

    If any tensor is floating-point, they will all be casted to the most-precise
    floating-point dtype. Otherwise the tensors are not casted.

    Args:
        tensors: A list of tensors.

    Returns:
        Same list, casted to a common dtype.
    """
    highest_float = None
    for x in tensors:
        dtype = backend.standardize_dtype(x.dtype)
        if is_float(dtype):
            if highest_float is None or dtype_size(dtype) > highest_float:
                highest_float = dtype
            elif dtype == "float16" and highest_float == "bfloat16":
                highest_float = "float32"
    if highest_float:
        tensors = [ops.cast(x, highest_float) for x in tensors]
    return tensors




"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""


from keras_core.src.initializers import deserialize
from keras_core.src.initializers import get
from keras_core.src.initializers import serialize
from keras_core.src.initializers.constant_initializers import Constant
from keras_core.src.initializers.constant_initializers import Constant as constant
from keras_core.src.initializers.constant_initializers import Identity
from keras_core.src.initializers.constant_initializers import Identity as IdentityInitializer
from keras_core.src.initializers.constant_initializers import Identity as identity
from keras_core.src.initializers.constant_initializers import Ones
from keras_core.src.initializers.constant_initializers import Ones as ones
from keras_core.src.initializers.constant_initializers import Zeros
from keras_core.src.initializers.constant_initializers import Zeros as zeros
from keras_core.src.initializers.initializer import Initializer
from keras_core.src.initializers.random_initializers import GlorotNormal
from keras_core.src.initializers.random_initializers import GlorotNormal as glorot_normal
from keras_core.src.initializers.random_initializers import GlorotUniform
from keras_core.src.initializers.random_initializers import GlorotUniform as glorot_uniform
from keras_core.src.initializers.random_initializers import HeNormal
from keras_core.src.initializers.random_initializers import HeNormal as he_normal
from keras_core.src.initializers.random_initializers import HeUniform
from keras_core.src.initializers.random_initializers import HeUniform as he_uniform
from keras_core.src.initializers.random_initializers import LecunNormal
from keras_core.src.initializers.random_initializers import LecunNormal as lecun_normal
from keras_core.src.initializers.random_initializers import LecunUniform
from keras_core.src.initializers.random_initializers import LecunUniform as lecun_uniform
from keras_core.src.initializers.random_initializers import OrthogonalInitializer
from keras_core.src.initializers.random_initializers import OrthogonalInitializer as Orthogonal
from keras_core.src.initializers.random_initializers import OrthogonalInitializer as orthogonal
from keras_core.src.initializers.random_initializers import RandomNormal
from keras_core.src.initializers.random_initializers import RandomNormal as random_normal
from keras_core.src.initializers.random_initializers import RandomUniform
from keras_core.src.initializers.random_initializers import RandomUniform as random_uniform
from keras_core.src.initializers.random_initializers import TruncatedNormal
from keras_core.src.initializers.random_initializers import TruncatedNormal as truncated_normal
from keras_core.src.initializers.random_initializers import VarianceScaling
from keras_core.src.initializers.random_initializers import VarianceScaling as variance_scaling



from keras_core.src.api_export import keras_core_export
from keras_core.src.optimizers.adadelta import Adadelta
from keras_core.src.optimizers.adafactor import Adafactor
from keras_core.src.optimizers.adagrad import Adagrad
from keras_core.src.optimizers.adam import Adam
from keras_core.src.optimizers.adamax import Adamax
from keras_core.src.optimizers.adamw import AdamW
from keras_core.src.optimizers.ftrl import Ftrl
from keras_core.src.optimizers.nadam import Nadam
from keras_core.src.optimizers.optimizer import Optimizer
from keras_core.src.optimizers.rmsprop import RMSprop
from keras_core.src.optimizers.sgd import SGD
from keras_core.src.saving import serialization_lib

ALL_OBJECTS = {
    Optimizer,
    Adam,
    SGD,
    RMSprop,
    Adadelta,
    AdamW,
    Adagrad,
    Adamax,
    Adafactor,
    Nadam,
    Ftrl,
}
ALL_OBJECTS_DICT = {cls.__name__.lower(): cls for cls in ALL_OBJECTS}


@keras_core_export("keras_core.optimizers.serialize")
def serialize(optimizer):
    """Returns the optimizer configuration as a Python dict.

    Args:
        optimizer: An `Optimizer` instance to serialize.

    Returns:
        Python dict which contains the configuration of the optimizer.
    """
    return serialization_lib.serialize_keras_object(optimizer)


@keras_core_export("keras_core.optimizers.deserialize")
def deserialize(config, custom_objects=None):
    """Returns a Keras optimizer object via its configuration.

    Args:
        config: Optimizer configuration dictionary.
        custom_objects: Optional dictionary mapping names (strings) to custom
            objects (classes and functions) to be considered during
            deserialization.

    Returns:
        A Keras Optimizer instance.
    """
    # Make deserialization case-insensitive for built-in optimizers.
    if config["class_name"].lower() in ALL_OBJECTS_DICT:
        config["class_name"] = config["class_name"].lower()

    return serialization_lib.deserialize_keras_object(
        config,
        module_objects=ALL_OBJECTS_DICT,
        custom_objects=custom_objects,
    )


@keras_core_export("keras_core.optimizers.get")
def get(identifier):
    """Retrieves a Keras Optimizer instance.

    Args:
        identifier: Optimizer identifier, one of:
            - String: name of an optimizer
            - Dictionary: configuration dictionary.
            - Keras Optimizer instance (it will be returned unchanged).

    Returns:
        A Keras Optimizer instance.
    """
    if identifier is None:
        return None
    elif isinstance(identifier, dict):
        obj = deserialize(identifier)
    elif isinstance(identifier, str):
        config = {"class_name": identifier, "config": {}}
        obj = deserialize(config)
    else:
        obj = identifier

    if isinstance(obj, Optimizer):
        return obj
    raise ValueError(f"Could not interpret optimizer identifier: {identifier}")




"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""


from keras_core.src.applications.efficientnet_v2 import EfficientNetV2B0
from keras_core.src.applications.efficientnet_v2 import EfficientNetV2B1
from keras_core.src.applications.efficientnet_v2 import EfficientNetV2B2
from keras_core.src.applications.efficientnet_v2 import EfficientNetV2B3
from keras_core.src.applications.efficientnet_v2 import EfficientNetV2L
from keras_core.src.applications.efficientnet_v2 import EfficientNetV2M
from keras_core.src.applications.efficientnet_v2 import EfficientNetV2S
from keras_core.src.applications.efficientnet_v2 import decode_predictions
from keras_core.src.applications.efficientnet_v2 import preprocess_input



import jax
import jax.numpy as jnp


def segment_sum(data, segment_ids, num_segments=None, sorted=False):
    if num_segments is None:
        raise ValueError(
            "Argument `num_segments` must be set when using the JAX backend. "
            "Received: num_segments=None"
        )
    return jax.ops.segment_sum(
        data, segment_ids, num_segments, indices_are_sorted=sorted
    )


def top_k(x, k, sorted=True):
    # Jax does not supported `sorted`, but in the case where `sorted=False`,
    # order is not guaranteed, so OK to return sorted output.
    return jax.lax.top_k(x, k)


def in_top_k(targets, predictions, k):
    targets = targets[..., None]
    topk_values = top_k(predictions, k)[0]
    targets_values = jnp.take_along_axis(predictions, targets, axis=-1)
    mask = targets_values >= topk_values
    return jax.numpy.any(mask, axis=1)


def logsumexp(x, axis=None, keepdims=False):
    max_x = jnp.max(x, axis=axis, keepdims=True)
    result = (
        jnp.log(jnp.sum(jnp.exp(x - max_x), axis=axis, keepdims=True)) + max_x
    )
    return jnp.squeeze(result) if not keepdims else result


def qr(x, mode="reduced"):
    if mode not in {"reduced", "complete"}:
        raise ValueError(
            "`mode` argument value not supported. "
            "Expected one of {'reduced', 'complete'}. "
            f"Received: mode={mode}"
        )
    return jax.numpy.linalg.qr(x, mode=mode)


def _get_complex_tensor_from_tuple(a):
    if not isinstance(a, (tuple, list)) or len(a) != 2:
        raise ValueError(
            "Input `a` should be a tuple of two tensors - real and imaginary."
            f"Received: a={a}"
        )
    # `convert_to_tensor` does not support passing complex tensors. We separate
    # the input out into real and imaginary and convert them separately.
    real, imag = a
    # Check shapes.
    if real.shape != imag.shape:
        raise ValueError(
            "Input `a` should be a tuple of two tensors - real and imaginary."
            "Both the real and imaginary parts should have the same shape. "
            f"Received: a[0].shape = {real.shape}, a[1].shape = {imag.shape}"
        )
    # Ensure dtype is float.
    if not jnp.issubdtype(real.dtype, jnp.floating) or not jnp.issubdtype(
        imag.dtype, jnp.floating
    ):
        raise ValueError(
            "At least one tensor in input `a` is not of type float."
            f"Received: a={a}."
        )
    complex_input = jax.lax.complex(real, imag)
    return complex_input


def fft(a):
    complex_input = _get_complex_tensor_from_tuple(a)
    complex_output = jax.numpy.fft.fft(complex_input)
    return jax.numpy.real(complex_output), jax.numpy.imag(complex_output)


def fft2(a):
    complex_input = _get_complex_tensor_from_tuple(a)
    complex_output = jax.numpy.fft.fft2(complex_input)
    return jax.numpy.real(complex_output), jax.numpy.imag(complex_output)




import types

import numpy as np
import tensorflow as tf
from tensorflow.compiler.tf2xla.python.xla import dynamic_update_slice

from keras_core.src.backend.common import KerasVariable
from keras_core.src.backend.common import standardize_dtype
from keras_core.src.backend.common.keras_tensor import KerasTensor
from keras_core.src.backend.common.stateless_scope import StatelessScope
from keras_core.src.utils.naming import auto_name

DYNAMIC_SHAPES_OK = True


class Variable(
    KerasVariable,
    tf.__internal__.types.Tensor,
    tf.__internal__.tracking.Trackable,
):
    _should_act_as_resource_variable = True

    @property
    def handle(self):
        return self.value.handle

    def _initialize(self, value):
        self._value = tf.Variable(
            value, dtype=self._dtype, trainable=self.trainable, name=self.name
        )

    def _direct_assign(self, value):
        self._value.assign(tf.cast(value, self._value.dtype))

    def _convert_to_tensor(self, value, dtype=None):
        return convert_to_tensor(value, dtype=dtype)

    def numpy(self):  # noqa: F811
        return self.value.numpy()

    @property
    def shape(self):
        return tf.TensorShape(super().shape)

    # Overload native accessor.
    def __tf_tensor__(self, dtype=None, name=None):
        return tf.convert_to_tensor(self.value, dtype=dtype, name=name)

    # Methods below are for SavedModel support
    @property
    def _shared_name(self):
        return self.value._shared_name

    def _serialize_to_tensors(self):
        return self.value._serialize_to_tensors()

    def _restore_from_tensors(self, restored_tensors):
        return self.value._restore_from_tensors(restored_tensors)

    def _export_to_saved_model_graph(
        self, object_map, tensor_map, options, **kwargs
    ):
        resource_list = self.value._export_to_saved_model_graph(
            object_map, tensor_map, options, **kwargs
        )
        object_map[self] = tf.Variable(object_map[self.value])
        return resource_list

    def _write_object_proto(self, proto, options):
        return self.value._write_object_proto(proto, options)


def convert_to_tensor(x, dtype=None):
    if dtype is not None:
        dtype = standardize_dtype(dtype)
        if tf.is_tensor(x):
            return tf.cast(x, dtype=dtype)
    return tf.convert_to_tensor(x, dtype=dtype)


def convert_to_numpy(x):
    return np.array(x)


def is_tensor(x):
    return tf.is_tensor(x)


def shape(x):
    return tf.shape(x)


def cast(x, dtype):
    dtype = standardize_dtype(dtype)
    return tf.cast(x, dtype=dtype)


def name_scope(name):
    return tf.name_scope(name)


def compute_output_spec(fn, *args, **kwargs):
    with StatelessScope():
        graph_name = auto_name("scratch_graph")
        with tf.__internal__.FuncGraph(graph_name).as_default():

            def convert_keras_tensor_to_tf(x):
                if isinstance(x, KerasTensor):
                    return tf.compat.v1.placeholder(
                        shape=x.shape, dtype=x.dtype
                    )
                if isinstance(x, types.FunctionType):

                    def _fn(*x_args, **x_kwargs):
                        out = x(*x_args, **x_kwargs)
                        out = convert_keras_tensor_to_tf(out)
                        return out

                    return _fn
                return x

            args, kwargs = tf.nest.map_structure(
                convert_keras_tensor_to_tf, (args, kwargs)
            )
            tf_out = fn(*args, **kwargs)

            def convert_tf_to_keras_tensor(x):
                if tf.is_tensor(x):
                    return KerasTensor(x.shape, x.dtype)
                return x

            output_spec = tf.nest.map_structure(
                convert_tf_to_keras_tensor, tf_out
            )
    return output_spec


def cond(pred, true_fn, false_fn):
    return tf.cond(pred, true_fn=true_fn, false_fn=false_fn)


def vectorized_map(function, elements):
    return tf.vectorized_map(function, elements)


def scatter(indices, values, shape):
    return tf.scatter_nd(indices, values, shape)


def scatter_update(inputs, indices, updates):
    return tf.tensor_scatter_nd_update(inputs, indices, updates)


def slice(inputs, start_indices, shape):
    return tf.slice(inputs, start_indices, shape)


def slice_update(inputs, start_indices, updates):
    return dynamic_update_slice(inputs, updates, start_indices)


def while_loop(
    cond,
    body,
    loop_vars,
    maximum_iterations=None,
):
    return tf.while_loop(
        cond,
        body,
        loop_vars,
        maximum_iterations=maximum_iterations,
    )


def fori_loop(lower, upper, body_fun, init_val):
    return tf.while_loop(
        lambda i, val: i < upper,
        lambda i, val: (i + 1, body_fun(i, val)),
        (lower, init_val),
    )[1]


def stop_gradient(variable):
    return tf.stop_gradient(variable)


def unstack(x, num=None, axis=0):
    return tf.unstack(x, num=num, axis=axis)




class NumpyTrainer:
    def fit(self):
        raise NotImplementedError("Trainer not implemented for NumPy backend.")

    def predict(self):
        raise NotImplementedError("Trainer not implemented for NumPy backend.")

    def evaluate(self):
        raise NotImplementedError("Trainer not implemented for NumPy backend.")

    def train_on_batch(self):
        raise NotImplementedError("Trainer not implemented for NumPy backend.")

    def test_on_batch(self):
        raise NotImplementedError("Trainer not implemented for NumPy backend.")

    def predict_on_batch(self):
        raise NotImplementedError("Trainer not implemented for NumPy backend.")




"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""


from keras_core.src.layers import deserialize
from keras_core.src.layers import serialize
from keras_core.src.layers.activations.activation import Activation
from keras_core.src.layers.activations.elu import ELU
from keras_core.src.layers.activations.leaky_relu import LeakyReLU
from keras_core.src.layers.activations.prelu import PReLU
from keras_core.src.layers.activations.relu import ReLU
from keras_core.src.layers.activations.softmax import Softmax
from keras_core.src.layers.attention.additive_attention import AdditiveAttention
from keras_core.src.layers.attention.attention import Attention
from keras_core.src.layers.attention.multi_head_attention import MultiHeadAttention
from keras_core.src.layers.convolutional.conv1d import Conv1D
from keras_core.src.layers.convolutional.conv1d import Conv1D as Convolution1D
from keras_core.src.layers.convolutional.conv1d_transpose import Conv1DTranspose
from keras_core.src.layers.convolutional.conv1d_transpose import Conv1DTranspose as Convolution1DTranspose
from keras_core.src.layers.convolutional.conv2d import Conv2D
from keras_core.src.layers.convolutional.conv2d import Conv2D as Convolution2D
from keras_core.src.layers.convolutional.conv2d_transpose import Conv2DTranspose
from keras_core.src.layers.convolutional.conv2d_transpose import Conv2DTranspose as Convolution2DTranspose
from keras_core.src.layers.convolutional.conv3d import Conv3D
from keras_core.src.layers.convolutional.conv3d import Conv3D as Convolution3D
from keras_core.src.layers.convolutional.conv3d_transpose import Conv3DTranspose
from keras_core.src.layers.convolutional.conv3d_transpose import Conv3DTranspose as Convolution3DTranspose
from keras_core.src.layers.convolutional.depthwise_conv1d import DepthwiseConv1D
from keras_core.src.layers.convolutional.depthwise_conv2d import DepthwiseConv2D
from keras_core.src.layers.convolutional.separable_conv1d import SeparableConv1D
from keras_core.src.layers.convolutional.separable_conv1d import SeparableConv1D as SeparableConvolution1D
from keras_core.src.layers.convolutional.separable_conv2d import SeparableConv2D
from keras_core.src.layers.convolutional.separable_conv2d import SeparableConv2D as SeparableConvolution2D
from keras_core.src.layers.core.dense import Dense
from keras_core.src.layers.core.einsum_dense import EinsumDense
from keras_core.src.layers.core.embedding import Embedding
from keras_core.src.layers.core.identity import Identity
from keras_core.src.layers.core.input_layer import Input
from keras_core.src.layers.core.input_layer import InputLayer
from keras_core.src.layers.core.lambda_layer import Lambda
from keras_core.src.layers.core.masking import Masking
from keras_core.src.layers.core.wrapper import Wrapper
from keras_core.src.layers.input_spec import InputSpec
from keras_core.src.layers.layer import Layer
from keras_core.src.layers.merging.add import Add
from keras_core.src.layers.merging.add import add
from keras_core.src.layers.merging.average import Average
from keras_core.src.layers.merging.average import average
from keras_core.src.layers.merging.concatenate import Concatenate
from keras_core.src.layers.merging.concatenate import concatenate
from keras_core.src.layers.merging.dot import Dot
from keras_core.src.layers.merging.dot import dot
from keras_core.src.layers.merging.maximum import Maximum
from keras_core.src.layers.merging.maximum import maximum
from keras_core.src.layers.merging.minimum import Minimum
from keras_core.src.layers.merging.minimum import minimum
from keras_core.src.layers.merging.multiply import Multiply
from keras_core.src.layers.merging.multiply import multiply
from keras_core.src.layers.merging.subtract import Subtract
from keras_core.src.layers.merging.subtract import subtract
from keras_core.src.layers.normalization.batch_normalization import BatchNormalization
from keras_core.src.layers.normalization.group_normalization import GroupNormalization
from keras_core.src.layers.normalization.layer_normalization import LayerNormalization
from keras_core.src.layers.normalization.spectral_normalization import SpectralNormalization
from keras_core.src.layers.normalization.unit_normalization import UnitNormalization
from keras_core.src.layers.pooling.average_pooling1d import AveragePooling1D
from keras_core.src.layers.pooling.average_pooling1d import AveragePooling1D as AvgPool1D
from keras_core.src.layers.pooling.average_pooling2d import AveragePooling2D
from keras_core.src.layers.pooling.average_pooling2d import AveragePooling2D as AvgPool2D
from keras_core.src.layers.pooling.average_pooling3d import AveragePooling3D
from keras_core.src.layers.pooling.average_pooling3d import AveragePooling3D as AvgPool3D
from keras_core.src.layers.pooling.global_average_pooling1d import GlobalAveragePooling1D
from keras_core.src.layers.pooling.global_average_pooling1d import GlobalAveragePooling1D as GlobalAvgPool1D
from keras_core.src.layers.pooling.global_average_pooling2d import GlobalAveragePooling2D
from keras_core.src.layers.pooling.global_average_pooling2d import GlobalAveragePooling2D as GlobalAvgPool2D
from keras_core.src.layers.pooling.global_average_pooling3d import GlobalAveragePooling3D
from keras_core.src.layers.pooling.global_average_pooling3d import GlobalAveragePooling3D as GlobalAvgPool3D
from keras_core.src.layers.pooling.global_max_pooling1d import GlobalMaxPooling1D
from keras_core.src.layers.pooling.global_max_pooling1d import GlobalMaxPooling1D as GlobalMaxPool1D
from keras_core.src.layers.pooling.global_max_pooling2d import GlobalMaxPooling2D
from keras_core.src.layers.pooling.global_max_pooling2d import GlobalMaxPooling2D as GlobalMaxPool2D
from keras_core.src.layers.pooling.global_max_pooling3d import GlobalMaxPooling3D
from keras_core.src.layers.pooling.global_max_pooling3d import GlobalMaxPooling3D as GlobalMaxPool3D
from keras_core.src.layers.pooling.max_pooling1d import MaxPooling1D
from keras_core.src.layers.pooling.max_pooling1d import MaxPooling1D as MaxPool1D
from keras_core.src.layers.pooling.max_pooling2d import MaxPooling2D
from keras_core.src.layers.pooling.max_pooling2d import MaxPooling2D as MaxPool2D
from keras_core.src.layers.pooling.max_pooling3d import MaxPooling3D
from keras_core.src.layers.pooling.max_pooling3d import MaxPooling3D as MaxPool3D
from keras_core.src.layers.preprocessing.category_encoding import CategoryEncoding
from keras_core.src.layers.preprocessing.center_crop import CenterCrop
from keras_core.src.layers.preprocessing.discretization import Discretization
from keras_core.src.layers.preprocessing.hashed_crossing import HashedCrossing
from keras_core.src.layers.preprocessing.hashing import Hashing
from keras_core.src.layers.preprocessing.integer_lookup import IntegerLookup
from keras_core.src.layers.preprocessing.normalization import Normalization
from keras_core.src.layers.preprocessing.random_brightness import RandomBrightness
from keras_core.src.layers.preprocessing.random_contrast import RandomContrast
from keras_core.src.layers.preprocessing.random_crop import RandomCrop
from keras_core.src.layers.preprocessing.random_flip import RandomFlip
from keras_core.src.layers.preprocessing.random_rotation import RandomRotation
from keras_core.src.layers.preprocessing.random_translation import RandomTranslation
from keras_core.src.layers.preprocessing.random_zoom import RandomZoom
from keras_core.src.layers.preprocessing.rescaling import Rescaling
from keras_core.src.layers.preprocessing.resizing import Resizing
from keras_core.src.layers.preprocessing.string_lookup import StringLookup
from keras_core.src.layers.preprocessing.text_vectorization import TextVectorization
from keras_core.src.layers.regularization.activity_regularization import ActivityRegularization
from keras_core.src.layers.regularization.dropout import Dropout
from keras_core.src.layers.regularization.gaussian_dropout import GaussianDropout
from keras_core.src.layers.regularization.gaussian_noise import GaussianNoise
from keras_core.src.layers.regularization.spatial_dropout import SpatialDropout1D
from keras_core.src.layers.regularization.spatial_dropout import SpatialDropout2D
from keras_core.src.layers.regularization.spatial_dropout import SpatialDropout3D
from keras_core.src.layers.reshaping.cropping1d import Cropping1D
from keras_core.src.layers.reshaping.cropping2d import Cropping2D
from keras_core.src.layers.reshaping.cropping3d import Cropping3D
from keras_core.src.layers.reshaping.flatten import Flatten
from keras_core.src.layers.reshaping.permute import Permute
from keras_core.src.layers.reshaping.repeat_vector import RepeatVector
from keras_core.src.layers.reshaping.reshape import Reshape
from keras_core.src.layers.reshaping.up_sampling1d import UpSampling1D
from keras_core.src.layers.reshaping.up_sampling2d import UpSampling2D
from keras_core.src.layers.reshaping.up_sampling3d import UpSampling3D
from keras_core.src.layers.reshaping.zero_padding1d import ZeroPadding1D
from keras_core.src.layers.reshaping.zero_padding2d import ZeroPadding2D
from keras_core.src.layers.reshaping.zero_padding3d import ZeroPadding3D
from keras_core.src.layers.rnn.bidirectional import Bidirectional
from keras_core.src.layers.rnn.conv_lstm1d import ConvLSTM1D
from keras_core.src.layers.rnn.conv_lstm2d import ConvLSTM2D
from keras_core.src.layers.rnn.conv_lstm3d import ConvLSTM3D
from keras_core.src.layers.rnn.gru import GRU
from keras_core.src.layers.rnn.gru import GRUCell
from keras_core.src.layers.rnn.lstm import LSTM
from keras_core.src.layers.rnn.lstm import LSTMCell
from keras_core.src.layers.rnn.rnn import RNN
from keras_core.src.layers.rnn.simple_rnn import SimpleRNN
from keras_core.src.layers.rnn.simple_rnn import SimpleRNNCell
from keras_core.src.layers.rnn.stacked_rnn_cells import StackedRNNCells
from keras_core.src.layers.rnn.time_distributed import TimeDistributed



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""


from keras_core.src.ops.image import affine_transform
from keras_core.src.ops.image import extract_patches
from keras_core.src.ops.image import resize



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""


from keras_core.src.applications.nasnet import NASNetLarge
from keras_core.src.applications.nasnet import NASNetMobile
from keras_core.src.applications.nasnet import decode_predictions
from keras_core.src.applications.nasnet import preprocess_input



"""
Separation of concerns:

DataAdapter:
    - x, y
    - sample_weight
    - class_weight
    - shuffle
    - batch_size
        - steps, as it relates to batch_size for array data

EpochIterator:
    - whether to yield numpy or tf data
    - steps
    - most argument validation

Trainer:
    - steps_per_execution
    - validation_split
    - validation_data
    - callbacks
    - validation_freq
    - epochs
    - initial_epoch
    - any backend-specific concern such as distribution

PyDataset:
    - num_workers
    - use_multiprocessing
    - max_queue_size

EpochIterator steps:

1. Look at data type and select correct DataHandler
2. Instantiate DataHandler with correct arguments
3. Raise or warn on unused arguments
4. in __iter__, iterate, either for a fixed number of steps
or until there is no data

"""
import types
import warnings

from keras_core.src.trainers.data_adapters import array_data_adapter
from keras_core.src.trainers.data_adapters import generator_data_adapter
from keras_core.src.trainers.data_adapters import py_dataset_adapter
from keras_core.src.trainers.data_adapters import tf_dataset_adapter
from keras_core.src.trainers.data_adapters import torch_data_adapter
from keras_core.src.utils.module_utils import tensorflow as tf


class EpochIterator:
    def __init__(
        self,
        x,
        y=None,
        sample_weight=None,
        batch_size=None,
        steps_per_epoch=None,
        shuffle=False,
        class_weight=None,
        steps_per_execution=1,
    ):
        self.steps_per_epoch = steps_per_epoch
        self.steps_per_execution = steps_per_execution
        if steps_per_epoch:
            self._current_iterator = None
            self._insufficient_data = False
        if array_data_adapter.can_convert_arrays((x, y, sample_weight)):
            self.data_adapter = array_data_adapter.ArrayDataAdapter(
                x,
                y,
                sample_weight=sample_weight,
                class_weight=class_weight,
                shuffle=shuffle,
                batch_size=batch_size,
                steps=steps_per_epoch,
            )
        elif tf.available and isinstance(x, tf.data.Dataset):
            self.data_adapter = tf_dataset_adapter.TFDatasetAdapter(
                x, class_weight=class_weight
            )
            # Unsupported args: y, sample_weight, shuffle
            if y is not None:
                raise_unsupported_arg("y", "the targets", "tf.data.Dataset")
            if sample_weight is not None:
                raise_unsupported_arg(
                    "sample_weights", "the sample weights", "tf.data.Dataset"
                )
            # TODO: should we warn or not?
            # warnings.warn(
            #     "`shuffle=True` was passed, but will be ignored since the "
            #     "data `x` was provided as a tf.data.Dataset. The Dataset is "
            #     "expected to already be shuffled "
            #     "(via `.shuffle(tf.data.AUTOTUNE)`)"
            # )
        elif isinstance(x, py_dataset_adapter.PyDataset):
            self.data_adapter = py_dataset_adapter.PyDatasetAdapter(
                x, class_weight=class_weight, shuffle=shuffle
            )
            if y is not None:
                raise_unsupported_arg("y", "the targets", "PyDataset")
            if sample_weight is not None:
                raise_unsupported_arg(
                    "sample_weights", "the sample weights", "PyDataset"
                )
        elif is_torch_dataloader(x):
            self.data_adapter = torch_data_adapter.TorchDataLoaderAdapter(x)
            if y is not None:
                raise_unsupported_arg("y", "the targets", "torch DataLoader")
            if sample_weight is not None:
                raise_unsupported_arg(
                    "sample_weights", "the sample weights", "torch DataLoader"
                )
            if class_weight is not None:
                raise ValueError(
                    "Argument `class_weight` is not supported for torch "
                    f"DataLoader inputs. Received: class_weight={class_weight}"
                )
            # TODO: should we warn or not?
            # warnings.warn(
            #     "`shuffle=True` was passed, but will be ignored since the "
            #     "data `x` was provided as a torch DataLoader. The DataLoader "
            #     "is expected to already be shuffled."
            # )
        elif isinstance(x, types.GeneratorType):
            self.data_adapter = generator_data_adapter.GeneratorDataAdapter(x)
            if y is not None:
                raise_unsupported_arg("y", "the targets", "PyDataset")
            if sample_weight is not None:
                raise_unsupported_arg(
                    "sample_weights", "the sample weights", "PyDataset"
                )
            if class_weight is not None:
                raise ValueError(
                    "Argument `class_weight` is not supported for Python "
                    f"generator inputs. Received: class_weight={class_weight}"
                )
            if shuffle:
                raise ValueError(
                    "Argument `shuffle` is not supported for Python generator "
                    f"inputs. Received: shuffle={shuffle}"
                )
        else:
            raise ValueError(
                f"Unrecognized data type: x={x} (of type {type(x)})"
            )
        self._num_batches = self.data_adapter.num_batches

    def _get_iterator(self, return_type):
        if return_type not in ("np", "tf"):
            raise ValueError(
                "Argument `return_type` must be one of `{'np', 'tf'}`. "
                f"Received instead: return_type={return_type}"
            )
        if return_type == "np":
            iterator = self.data_adapter.get_numpy_iterator()
        else:
            iterator = self.data_adapter.get_tf_dataset()
        return iterator

    def enumerate_epoch(self, return_type="np"):
        buffer = []
        if self.steps_per_epoch:
            if not self._current_iterator:
                self._current_iterator = self._get_iterator(return_type)
                self._insufficient_data = False

            for step in range(self.steps_per_epoch):
                if self._insufficient_data:
                    break

                if tf.available:
                    errors = (StopIteration, tf.errors.OutOfRangeError)
                else:
                    errors = (StopIteration,)

                try:
                    data = next(self._current_iterator)
                    buffer.append(data)
                    if len(buffer) == self.steps_per_execution:
                        yield step - len(buffer) + 1, buffer
                        buffer = []
                except errors:
                    warnings.warn(
                        "Your input ran out of data; interrupting epoch. "
                        "Make sure that your dataset or generator can generate "
                        "at least `steps_per_epoch * epochs` batches. "
                        "You may need to use the `.repeat()` "
                        "function when building your dataset.",
                        stacklevel=2,
                    )
                    self._current_iterator = None
                    self._insufficient_data = True
            if buffer:
                yield step - len(buffer) + 1, buffer
        else:
            for step, data in enumerate(self._get_iterator(return_type)):
                buffer.append(data)
                if len(buffer) == self.steps_per_execution:
                    yield step - len(buffer) + 1, buffer
                    buffer = []
            if buffer:
                yield step - len(buffer) + 1, buffer
            if not self._num_batches:
                # Infer the number of batches returned by the data_adater.
                # Assumed static.
                self._num_batches = step + 1
        self.data_adapter.on_epoch_end()

    @property
    def num_batches(self):
        if self.steps_per_epoch:
            return self.steps_per_epoch
        # Either copied from the data_adapter, or
        # inferred at the end of an iteration.
        return self._num_batches


def raise_unsupported_arg(arg_name, arg_description, input_type):
    raise ValueError(
        f"When providing `x` as a {input_type}, `{arg_name}` "
        f"should not be passed. Instead, the {arg_description} should "
        f"be included as part of the {input_type}."
    )


def is_torch_dataloader(x):
    if hasattr(x, "__class__"):
        for parent in x.__class__.__mro__:
            if parent.__name__ == "DataLoader" and str(
                parent.__module__
            ).startswith("torch.utils.data"):
                return True
    return False




"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""


from keras_core.src.utils.image_utils import array_to_img
from keras_core.src.utils.image_utils import img_to_array
from keras_core.src.utils.image_utils import load_img
from keras_core.src.utils.image_utils import save_img

"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""


from keras_core.src.legacy.preprocessing.image import DirectoryIterator
from keras_core.src.legacy.preprocessing.image import ImageDataGenerator
from keras_core.src.legacy.preprocessing.image import Iterator
from keras_core.src.legacy.preprocessing.image import NumpyArrayIterator
from keras_core.src.legacy.preprocessing.image import apply_affine_transform
from keras_core.src.legacy.preprocessing.image import apply_brightness_shift
from keras_core.src.legacy.preprocessing.image import apply_channel_shift
from keras_core.src.legacy.preprocessing.image import random_brightness
from keras_core.src.legacy.preprocessing.image import random_channel_shift
from keras_core.src.legacy.preprocessing.image import random_rotation
from keras_core.src.legacy.preprocessing.image import random_shear
from keras_core.src.legacy.preprocessing.image import random_shift
from keras_core.src.legacy.preprocessing.image import random_zoom



import tensorflow as tf
from tensorflow.experimental import numpy as tfnp

from keras_core.src.backend.tensorflow.core import convert_to_tensor


def add(x1, x2):
    return tfnp.add(x1, x2)


def bincount(x, weights=None, minlength=None):
    if minlength is not None:
        x = tf.cast(x, tf.int32)
    return tf.math.bincount(x, weights=weights, minlength=minlength, axis=-1)


def einsum(subscripts, *operands, **kwargs):
    return tfnp.einsum(subscripts, *operands, **kwargs)


def subtract(x1, x2):
    return tfnp.subtract(x1, x2)


def matmul(x1, x2):
    return tfnp.matmul(x1, x2)


def multiply(x1, x2):
    return tfnp.multiply(x1, x2)


def mean(x, axis=None, keepdims=False):
    return tfnp.mean(x, axis=axis, keepdims=keepdims)


def max(x, axis=None, keepdims=False, initial=None):
    # The TensorFlow numpy API implementation doesn't support `initial` so we
    # handle it manually here.
    if initial is not None:
        return tf.math.maximum(
            tfnp.max(x, axis=axis, keepdims=keepdims), initial
        )

    # TensorFlow returns -inf by default for an empty list, but for consistency
    # with other backends and the numpy API we want to throw in this case.
    if tf.executing_eagerly():
        size_x = size(x)
        tf.assert_greater(
            size_x,
            tf.constant(0, dtype=size_x.dtype),
            message="Cannot compute the max of an empty tensor.",
        )

    return tfnp.max(x, axis=axis, keepdims=keepdims)


def ones(shape, dtype="float32"):
    return tf.ones(shape, dtype=dtype)


def zeros(shape, dtype="float32"):
    return tf.zeros(shape, dtype=dtype)


def absolute(x):
    return tfnp.absolute(x)


def abs(x):
    return absolute(x)


def all(x, axis=None, keepdims=False):
    return tfnp.all(x, axis=axis, keepdims=keepdims)


def any(x, axis=None, keepdims=False):
    return tfnp.any(x, axis=axis, keepdims=keepdims)


def amax(x, axis=None, keepdims=False):
    return tfnp.amax(x, axis=axis, keepdims=keepdims)


def amin(x, axis=None, keepdims=False):
    return tfnp.amin(x, axis=axis, keepdims=keepdims)


def append(
    x1,
    x2,
    axis=None,
):
    return tfnp.append(x1, x2, axis=axis)


def arange(start, stop=None, step=1, dtype=None):
    # tfnp.arange has trouble with dynamic Tensors in compiled function.
    # tf.range does not.
    return tf.range(start, stop, delta=step, dtype=dtype)


def arccos(x):
    return tfnp.arccos(x)


def arcsin(x):
    return tfnp.arcsin(x)


def arctan(x):
    return tfnp.arctan(x)


def arctan2(x1, x2):
    return tfnp.arctan2(x1, x2)


def argmax(x, axis=None):
    return tfnp.argmax(x, axis=axis)


def argmin(x, axis=None):
    return tfnp.argmin(x, axis=axis)


def argsort(x, axis=-1):
    return tfnp.argsort(x, axis=axis)


def array(x, dtype=None):
    return tfnp.array(x, dtype=dtype)


def average(x, axis=None, weights=None):
    if not isinstance(axis, (list, tuple)):
        axis = (axis,)
    for a in axis:
        # `tfnp.average` does not handle multiple axes.
        x = tfnp.average(x, weights=weights, axis=a)
    return x


def broadcast_to(x, shape):
    return tfnp.broadcast_to(x, shape)


def ceil(x):
    return tfnp.ceil(x)


def clip(x, x_min, x_max):
    return tfnp.clip(x, x_min, x_max)


def concatenate(xs, axis=0):
    return tfnp.concatenate(xs, axis=axis)


def conjugate(x):
    return tfnp.conjugate(x)


def conj(x):
    return conjugate(x)


def copy(x):
    return tfnp.copy(x)


def cos(x):
    return tfnp.cos(x)


def count_nonzero(x, axis=None):
    return tfnp.count_nonzero(x, axis=axis)


def cross(x1, x2, axisa=-1, axisb=-1, axisc=-1, axis=None):
    return tfnp.cross(
        x1,
        x2,
        axisa=axisa,
        axisb=axisb,
        axisc=axisc,
        axis=axis,
    )


def cumprod(x, axis=None):
    return tfnp.cumprod(x, axis=axis)


def cumsum(x, axis=None):
    return tfnp.cumsum(x, axis=axis)


def diag(x, k=0):
    return tfnp.diag(x, k=k)


def diagonal(x, offset=0, axis1=0, axis2=1):
    return tfnp.diagonal(
        x,
        offset=offset,
        axis1=axis1,
        axis2=axis2,
    )


def dot(x, y):
    return tfnp.dot(x, y)


def empty(shape, dtype="float32"):
    return tfnp.empty(shape, dtype=dtype)


def equal(x1, x2):
    return tfnp.equal(x1, x2)


def exp(x):
    return tfnp.exp(x)


def expand_dims(x, axis):
    return tfnp.expand_dims(x, axis)


def expm1(x):
    return tfnp.expm1(x)


def flip(x, axis=None):
    return tfnp.flip(x, axis=axis)


def floor(x):
    return tfnp.floor(x)


def full(shape, fill_value, dtype=None):
    return tfnp.full(shape, fill_value, dtype=dtype)


def full_like(x, fill_value, dtype=None):
    return tfnp.full_like(x, fill_value, dtype=dtype)


def greater(x1, x2):
    return tfnp.greater(x1, x2)


def greater_equal(x1, x2):
    return tfnp.greater_equal(x1, x2)


def hstack(xs):
    return tfnp.hstack(xs)


def identity(n, dtype="float32"):
    return tfnp.identity(n, dtype=dtype)


def imag(x):
    return tfnp.imag(x)


def isclose(x1, x2):
    return tfnp.isclose(x1, x2)


def isfinite(x):
    return tfnp.isfinite(x)


def isinf(x):
    return tfnp.isinf(x)


def isnan(x):
    return tfnp.isnan(x)


def less(x1, x2):
    return tfnp.less(x1, x2)


def less_equal(x1, x2):
    return tfnp.less_equal(x1, x2)


def linspace(
    start, stop, num=50, endpoint=True, retstep=False, dtype=None, axis=0
):
    return tfnp.linspace(
        start,
        stop,
        num=num,
        endpoint=endpoint,
        retstep=retstep,
        dtype=dtype,
        axis=axis,
    )


def log(x):
    return tfnp.log(x)


def log10(x):
    return tfnp.log10(x)


def log1p(x):
    return tfnp.log1p(x)


def log2(x):
    return tfnp.log2(x)


def logaddexp(x1, x2):
    return tfnp.logaddexp(x1, x2)


def logical_and(x1, x2):
    return tfnp.logical_and(x1, x2)


def logical_not(x):
    return tfnp.logical_not(x)


def logical_or(x1, x2):
    return tfnp.logical_or(x1, x2)


def logspace(start, stop, num=50, endpoint=True, base=10, dtype=None, axis=0):
    return tfnp.logspace(
        start,
        stop,
        num=num,
        endpoint=endpoint,
        base=base,
        dtype=dtype,
        axis=axis,
    )


def maximum(x1, x2):
    return tfnp.maximum(x1, x2)


def meshgrid(*x, indexing="xy"):
    return tfnp.meshgrid(*x, indexing=indexing)


def min(x, axis=None, keepdims=False, initial=None):
    # The TensorFlow numpy API implementation doesn't support `initial` so we
    # handle it manually here.
    if initial is not None:
        return tf.math.minimum(
            tfnp.min(x, axis=axis, keepdims=keepdims), initial
        )

    # TensorFlow returns inf by default for an empty list, but for consistency
    # with other backends and the numpy API we want to throw in this case.
    if tf.executing_eagerly():
        size_x = size(x)
        tf.assert_greater(
            size_x,
            tf.constant(0, dtype=size_x.dtype),
            message="Cannot compute the min of an empty tensor.",
        )

    return tfnp.min(x, axis=axis, keepdims=keepdims)


def minimum(x1, x2):
    return tfnp.minimum(x1, x2)


def mod(x1, x2):
    return tfnp.mod(x1, x2)


def moveaxis(x, source, destination):
    return tfnp.moveaxis(x, source=source, destination=destination)


def nan_to_num(x):
    # Replace NaN with 0
    x = tf.where(tf.math.is_nan(x), 0, x)

    # Replace positive infinitiy with dtype.max
    x = tf.where(tf.math.is_inf(x) & (x > 0), x.dtype.max, x)

    # Replace negative infinity with dtype.min
    x = tf.where(tf.math.is_inf(x) & (x < 0), x.dtype.min, x)

    return x


def ndim(x):
    return tfnp.ndim(x)


def nonzero(x):
    return tfnp.nonzero(x)


def not_equal(x1, x2):
    return tfnp.not_equal(x1, x2)


def ones_like(x, dtype=None):
    return tfnp.ones_like(x, dtype=dtype)


def zeros_like(x, dtype=None):
    return tf.zeros_like(x, dtype=dtype)


def outer(x1, x2):
    return tfnp.outer(x1, x2)


def pad(x, pad_width, mode="constant"):
    return tfnp.pad(x, pad_width, mode=mode)


def prod(x, axis=None, keepdims=False, dtype=None):
    return tfnp.prod(x, axis=axis, keepdims=keepdims, dtype=dtype)


def ravel(x):
    return tfnp.ravel(x)


def real(x):
    return tfnp.real(x)


def reciprocal(x):
    return tfnp.reciprocal(x)


def repeat(x, repeats, axis=None):
    # tfnp.repeat has trouble with dynamic Tensors in compiled function.
    # tf.repeat does not.
    return tf.repeat(x, repeats, axis=axis)


def reshape(x, new_shape):
    return tfnp.reshape(x, new_shape)


def roll(x, shift, axis=None):
    return tfnp.roll(x, shift, axis=axis)


def sign(x):
    return tfnp.sign(x)


def sin(x):
    return tfnp.sin(x)


def size(x):
    return tfnp.size(x)


def sort(x, axis=-1):
    return tfnp.sort(x, axis=axis)


def split(x, indices_or_sections, axis=0):
    return tfnp.split(x, indices_or_sections, axis=axis)


def stack(x, axis=0):
    return tfnp.stack(x, axis=axis)


def std(x, axis=None, keepdims=False):
    return tfnp.std(x, axis=axis, keepdims=keepdims)


def swapaxes(x, axis1, axis2):
    return tfnp.swapaxes(x, axis1=axis1, axis2=axis2)


def take(x, indices, axis=None):
    return tfnp.take(x, indices, axis=axis)


def take_along_axis(x, indices, axis=None):
    return tfnp.take_along_axis(x, indices, axis=axis)


def tan(x):
    return tfnp.tan(x)


def tensordot(x1, x2, axes=2):
    return tfnp.tensordot(x1, x2, axes=axes)


def round(x, decimals=0):
    return tfnp.round(x, decimals=decimals)


def tile(x, repeats):
    # The TFNP implementation is buggy, we roll our own.
    x = convert_to_tensor(x)
    repeats = tf.reshape(convert_to_tensor(repeats, dtype="int32"), [-1])
    repeats_size = tf.size(repeats)
    repeats = tf.pad(
        repeats,
        [[tf.maximum(x.shape.rank - repeats_size, 0), 0]],
        constant_values=1,
    )
    x_shape = tf.pad(
        tf.shape(x),
        [[tf.maximum(repeats_size - x.shape.rank, 0), 0]],
        constant_values=1,
    )
    x = tf.reshape(x, x_shape)
    return tf.tile(x, repeats)


def trace(x, offset=0, axis1=0, axis2=1):
    return tfnp.trace(x, offset=offset, axis1=axis1, axis2=axis2)


def tri(N, M=None, k=0, dtype="float32"):
    return tfnp.tri(N, M=M, k=k, dtype=dtype)


def tril(x, k=0):
    return tfnp.tril(x, k=k)


def triu(x, k=0):
    return tfnp.triu(x, k=k)


def vdot(x1, x2):
    return tfnp.vdot(x1, x2)


def vstack(xs):
    return tfnp.vstack(xs)


def where(condition, x1, x2):
    return tfnp.where(condition, x1, x2)


def divide(x1, x2):
    return tfnp.divide(x1, x2)


def true_divide(x1, x2):
    return tfnp.true_divide(x1, x2)


def power(x1, x2):
    return tfnp.power(x1, x2)


def negative(x):
    return tfnp.negative(x)


def square(x):
    return tfnp.square(x)


def sqrt(x):
    return tfnp.sqrt(x)


def squeeze(x, axis=None):
    return tfnp.squeeze(x, axis=axis)


def transpose(x, axes=None):
    return tfnp.transpose(x, axes=axes)


def var(x, axis=None, keepdims=False):
    return tfnp.var(x, axis=axis, keepdims=keepdims)


def sum(x, axis=None, keepdims=False):
    return tfnp.sum(x, axis=axis, keepdims=keepdims)


def eye(N, M=None, k=0, dtype="float32"):
    return tfnp.eye(N, M=M, k=k, dtype=dtype)


def floor_divide(x1, x2):
    return tfnp.floor_divide(x1, x2)


def logical_xor(x1, x2):
    return tfnp.logical_xor(x1, x2)




import tensorflow as tf

RESIZE_INTERPOLATIONS = (
    "bilinear",
    "nearest",
    "lanczos3",
    "lanczos5",
    "bicubic",
)


def resize(
    image,
    size,
    interpolation="bilinear",
    antialias=False,
    data_format="channels_last",
):
    if interpolation not in RESIZE_INTERPOLATIONS:
        raise ValueError(
            "Invalid value for argument `interpolation`. Expected of one "
            f"{RESIZE_INTERPOLATIONS}. Received: interpolation={interpolation}"
        )
    if not len(size) == 2:
        raise ValueError(
            "Argument `size` must be a tuple of two elements "
            f"(height, width). Received: size={size}"
        )
    size = tuple(size)
    if data_format == "channels_first":
        if len(image.shape) == 4:
            image = tf.transpose(image, (0, 2, 3, 1))
        elif len(image.shape) == 3:
            image = tf.transpose(image, (1, 2, 0))
        else:
            raise ValueError(
                "Invalid input rank: expected rank 3 (single image) "
                "or rank 4 (batch of images). Received input with shape: "
                f"image.shape={image.shape}"
            )

    resized = tf.image.resize(
        image, size, method=interpolation, antialias=antialias
    )
    if data_format == "channels_first":
        if len(image.shape) == 4:
            resized = tf.transpose(resized, (0, 3, 1, 2))
        elif len(image.shape) == 3:
            resized = tf.transpose(resized, (2, 0, 1))
    return resized


AFFINE_TRANSFORM_INTERPOLATIONS = (
    "nearest",
    "bilinear",
)
AFFINE_TRANSFORM_FILL_MODES = (
    "constant",
    "nearest",
    "wrap",
    # "mirror", not supported by TF
    "reflect",
)


def affine_transform(
    image,
    transform,
    interpolation="bilinear",
    fill_mode="constant",
    fill_value=0,
    data_format="channels_last",
):
    if interpolation not in AFFINE_TRANSFORM_INTERPOLATIONS:
        raise ValueError(
            "Invalid value for argument `interpolation`. Expected of one "
            f"{AFFINE_TRANSFORM_INTERPOLATIONS}. Received: "
            f"interpolation={interpolation}"
        )
    if fill_mode not in AFFINE_TRANSFORM_FILL_MODES:
        raise ValueError(
            "Invalid value for argument `fill_mode`. Expected of one "
            f"{AFFINE_TRANSFORM_FILL_MODES}. Received: fill_mode={fill_mode}"
        )
    if len(image.shape) not in (3, 4):
        raise ValueError(
            "Invalid image rank: expected rank 3 (single image) "
            "or rank 4 (batch of images). Received input with shape: "
            f"image.shape={image.shape}"
        )
    if len(transform.shape) not in (1, 2):
        raise ValueError(
            "Invalid transform rank: expected rank 1 (single transform) "
            "or rank 2 (batch of transforms). Received input with shape: "
            f"transform.shape={transform.shape}"
        )
    # unbatched case
    need_squeeze = False
    if len(image.shape) == 3:
        image = tf.expand_dims(image, axis=0)
        need_squeeze = True
    if len(transform.shape) == 1:
        transform = tf.expand_dims(transform, axis=0)

    if data_format == "channels_first":
        image = tf.transpose(image, (0, 2, 3, 1))

    affined = tf.raw_ops.ImageProjectiveTransformV3(
        images=image,
        transforms=tf.cast(transform, dtype=tf.float32),
        output_shape=tf.shape(image)[1:-1],
        fill_value=fill_value,
        interpolation=interpolation.upper(),
        fill_mode=fill_mode.upper(),
    )

    if data_format == "channels_first":
        affined = tf.transpose(affined, (0, 3, 1, 2))
    if need_squeeze:
        affined = tf.squeeze(affined, axis=0)
    return affined




"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""


from keras_core import activations
from keras_core import applications
from keras_core import callbacks
from keras_core import config
from keras_core import constraints
from keras_core import datasets
from keras_core import initializers
from keras_core import legacy
from keras_core import metrics
from keras_core import mixed_precision
from keras_core import models
from keras_core import ops
from keras_core import optimizers
from keras_core import random
from keras_core import regularizers
from keras_core import utils
from keras_core.src.backend.common.keras_tensor import KerasTensor
from keras_core.src.backend.common.stateless_scope import StatelessScope
from keras_core.src.initializers.initializer import Initializer
from keras_core.src.layers.core.input_layer import Input
from keras_core.src.layers.input_spec import InputSpec
from keras_core.src.layers.layer import Layer
from keras_core.src.losses.loss import Loss
from keras_core.src.metrics.metric import Metric
from keras_core.src.models.model import Model
from keras_core.src.models.sequential import Sequential
from keras_core.src.ops.function import Function
from keras_core.src.ops.operation import Operation
from keras_core.src.optimizers.optimizer import Optimizer
from keras_core.src.regularizers.regularizers import Regularizer

"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""


from keras_core._tf_keras import backend
from keras_core._tf_keras import layers
from keras_core._tf_keras import losses
from keras_core._tf_keras import metrics
from keras_core._tf_keras import preprocessing



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""


from keras_core.src.losses import deserialize
from keras_core.src.losses import get
from keras_core.src.losses import serialize
from keras_core.src.losses.loss import Loss
from keras_core.src.losses.losses import BinaryCrossentropy
from keras_core.src.losses.losses import BinaryFocalCrossentropy
from keras_core.src.losses.losses import CategoricalCrossentropy
from keras_core.src.losses.losses import CategoricalFocalCrossentropy
from keras_core.src.losses.losses import CategoricalHinge
from keras_core.src.losses.losses import CosineSimilarity
from keras_core.src.losses.losses import Hinge
from keras_core.src.losses.losses import Huber
from keras_core.src.losses.losses import KLDivergence
from keras_core.src.losses.losses import LogCosh
from keras_core.src.losses.losses import MeanAbsoluteError
from keras_core.src.losses.losses import MeanAbsolutePercentageError
from keras_core.src.losses.losses import MeanSquaredError
from keras_core.src.losses.losses import MeanSquaredLogarithmicError
from keras_core.src.losses.losses import Poisson
from keras_core.src.losses.losses import SparseCategoricalCrossentropy
from keras_core.src.losses.losses import SquaredHinge
from keras_core.src.losses.losses import binary_crossentropy
from keras_core.src.losses.losses import binary_focal_crossentropy
from keras_core.src.losses.losses import categorical_crossentropy
from keras_core.src.losses.losses import categorical_focal_crossentropy
from keras_core.src.losses.losses import categorical_hinge
from keras_core.src.losses.losses import cosine_similarity
from keras_core.src.losses.losses import hinge
from keras_core.src.losses.losses import huber
from keras_core.src.losses.losses import kl_divergence
from keras_core.src.losses.losses import log_cosh
from keras_core.src.losses.losses import mean_absolute_error
from keras_core.src.losses.losses import mean_absolute_percentage_error
from keras_core.src.losses.losses import mean_squared_error
from keras_core.src.losses.losses import mean_squared_logarithmic_error
from keras_core.src.losses.losses import poisson
from keras_core.src.losses.losses import sparse_categorical_crossentropy
from keras_core.src.losses.losses import squared_hinge



from keras_core.src import backend
from keras_core.src.api_export import keras_core_export
from keras_core.src.optimizers import base_optimizer

if backend.backend() == "tensorflow":
    from keras_core.src.backend.tensorflow.optimizer import TFOptimizer

    BackendOptimizer = TFOptimizer
elif backend.backend() == "torch":
    from keras_core.src.backend.torch.optimizers import TorchOptimizer

    BackendOptimizer = TorchOptimizer
else:
    BackendOptimizer = base_optimizer.BaseOptimizer


@keras_core_export(["keras_core.Optimizer", "keras_core.optimizers.Optimizer"])
class Optimizer(BackendOptimizer):
    pass


base_optimizer_keyword_args = base_optimizer.base_optimizer_keyword_args
Optimizer.__doc__ = base_optimizer.BaseOptimizer.__doc__




from keras_core.src import activations
from keras_core.src import applications
from keras_core.src import backend
from keras_core.src import constraints
from keras_core.src import datasets
from keras_core.src import initializers
from keras_core.src import layers
from keras_core.src import models
from keras_core.src import ops
from keras_core.src import optimizers
from keras_core.src import regularizers
from keras_core.src import utils
from keras_core.src.backend import KerasTensor
from keras_core.src.layers import Input
from keras_core.src.layers import Layer
from keras_core.src.models import Functional
from keras_core.src.models import Model
from keras_core.src.models import Sequential
from keras_core.src.version import __version__




"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""


from keras_core.src.constraints import deserialize
from keras_core.src.constraints import get
from keras_core.src.constraints import serialize
from keras_core.src.constraints.constraints import Constraint
from keras_core.src.constraints.constraints import MaxNorm
from keras_core.src.constraints.constraints import MaxNorm as max_norm
from keras_core.src.constraints.constraints import MinMaxNorm
from keras_core.src.constraints.constraints import MinMaxNorm as min_max_norm
from keras_core.src.constraints.constraints import NonNeg
from keras_core.src.constraints.constraints import NonNeg as non_neg
from keras_core.src.constraints.constraints import UnitNorm
from keras_core.src.constraints.constraints import UnitNorm as unit_norm



from keras_core.src.optimizers.schedules.learning_rate_schedule import CosineDecay
from keras_core.src.optimizers.schedules.learning_rate_schedule import (
    CosineDecayRestarts,
)
from keras_core.src.optimizers.schedules.learning_rate_schedule import (
    ExponentialDecay,
)
from keras_core.src.optimizers.schedules.learning_rate_schedule import (
    InverseTimeDecay,
)
from keras_core.src.optimizers.schedules.learning_rate_schedule import (
    PiecewiseConstantDecay,
)
from keras_core.src.optimizers.schedules.learning_rate_schedule import (
    PolynomialDecay,
)




"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""


from keras_core.src.regularizers import deserialize
from keras_core.src.regularizers import get
from keras_core.src.regularizers import serialize
from keras_core.src.regularizers.regularizers import L1
from keras_core.src.regularizers.regularizers import L1 as l1
from keras_core.src.regularizers.regularizers import L1L2
from keras_core.src.regularizers.regularizers import L1L2 as l1_l2
from keras_core.src.regularizers.regularizers import L2
from keras_core.src.regularizers.regularizers import L2 as l2
from keras_core.src.regularizers.regularizers import OrthogonalRegularizer
from keras_core.src.regularizers.regularizers import OrthogonalRegularizer as orthogonal_regularizer
from keras_core.src.regularizers.regularizers import Regularizer



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""


from keras_core._tf_keras.preprocessing import image
from keras_core._tf_keras.preprocessing import sequence
from keras_core.src.utils.dataset_utils import image_dataset_from_directory
from keras_core.src.utils.dataset_utils import text_dataset_from_directory
from keras_core.src.utils.dataset_utils import timeseries_dataset_from_array

"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""


from keras_core._tf_keras.preprocessing import image
from keras_core._tf_keras.preprocessing import sequence
from keras_core._tf_keras.preprocessing import text



from keras_core.src.api_export import keras_core_export
from keras_core.src.layers.activations.activation import Activation
from keras_core.src.layers.activations.elu import ELU
from keras_core.src.layers.activations.leaky_relu import LeakyReLU
from keras_core.src.layers.activations.prelu import PReLU
from keras_core.src.layers.activations.relu import ReLU
from keras_core.src.layers.activations.softmax import Softmax
from keras_core.src.layers.attention.additive_attention import AdditiveAttention
from keras_core.src.layers.attention.attention import Attention
from keras_core.src.layers.attention.multi_head_attention import MultiHeadAttention
from keras_core.src.layers.convolutional.conv1d import Conv1D
from keras_core.src.layers.convolutional.conv1d_transpose import Conv1DTranspose
from keras_core.src.layers.convolutional.conv2d import Conv2D
from keras_core.src.layers.convolutional.conv2d_transpose import Conv2DTranspose
from keras_core.src.layers.convolutional.conv3d import Conv3D
from keras_core.src.layers.convolutional.conv3d_transpose import Conv3DTranspose
from keras_core.src.layers.convolutional.depthwise_conv1d import DepthwiseConv1D
from keras_core.src.layers.convolutional.depthwise_conv2d import DepthwiseConv2D
from keras_core.src.layers.convolutional.separable_conv1d import SeparableConv1D
from keras_core.src.layers.convolutional.separable_conv2d import SeparableConv2D
from keras_core.src.layers.core.dense import Dense
from keras_core.src.layers.core.einsum_dense import EinsumDense
from keras_core.src.layers.core.embedding import Embedding
from keras_core.src.layers.core.identity import Identity
from keras_core.src.layers.core.input_layer import Input
from keras_core.src.layers.core.input_layer import InputLayer
from keras_core.src.layers.core.lambda_layer import Lambda
from keras_core.src.layers.core.masking import Masking
from keras_core.src.layers.core.wrapper import Wrapper
from keras_core.src.layers.layer import Layer
from keras_core.src.layers.merging.add import Add
from keras_core.src.layers.merging.add import add
from keras_core.src.layers.merging.average import Average
from keras_core.src.layers.merging.average import average
from keras_core.src.layers.merging.concatenate import Concatenate
from keras_core.src.layers.merging.concatenate import concatenate
from keras_core.src.layers.merging.dot import Dot
from keras_core.src.layers.merging.dot import dot
from keras_core.src.layers.merging.maximum import Maximum
from keras_core.src.layers.merging.maximum import maximum
from keras_core.src.layers.merging.minimum import Minimum
from keras_core.src.layers.merging.minimum import minimum
from keras_core.src.layers.merging.multiply import Multiply
from keras_core.src.layers.merging.multiply import multiply
from keras_core.src.layers.merging.subtract import Subtract
from keras_core.src.layers.merging.subtract import subtract
from keras_core.src.layers.normalization.batch_normalization import (
    BatchNormalization,
)
from keras_core.src.layers.normalization.group_normalization import (
    GroupNormalization,
)
from keras_core.src.layers.normalization.layer_normalization import (
    LayerNormalization,
)
from keras_core.src.layers.normalization.spectral_normalization import (
    SpectralNormalization,
)
from keras_core.src.layers.normalization.unit_normalization import UnitNormalization
from keras_core.src.layers.pooling.average_pooling1d import AveragePooling1D
from keras_core.src.layers.pooling.average_pooling2d import AveragePooling2D
from keras_core.src.layers.pooling.average_pooling3d import AveragePooling3D
from keras_core.src.layers.pooling.global_average_pooling1d import (
    GlobalAveragePooling1D,
)
from keras_core.src.layers.pooling.global_average_pooling2d import (
    GlobalAveragePooling2D,
)
from keras_core.src.layers.pooling.global_average_pooling3d import (
    GlobalAveragePooling3D,
)
from keras_core.src.layers.pooling.global_max_pooling1d import GlobalMaxPooling1D
from keras_core.src.layers.pooling.global_max_pooling2d import GlobalMaxPooling2D
from keras_core.src.layers.pooling.global_max_pooling3d import GlobalMaxPooling3D
from keras_core.src.layers.pooling.max_pooling1d import MaxPooling1D
from keras_core.src.layers.pooling.max_pooling2d import MaxPooling2D
from keras_core.src.layers.pooling.max_pooling3d import MaxPooling3D
from keras_core.src.layers.preprocessing.category_encoding import CategoryEncoding
from keras_core.src.layers.preprocessing.center_crop import CenterCrop
from keras_core.src.layers.preprocessing.discretization import Discretization
from keras_core.src.layers.preprocessing.hashed_crossing import HashedCrossing
from keras_core.src.layers.preprocessing.hashing import Hashing
from keras_core.src.layers.preprocessing.integer_lookup import IntegerLookup
from keras_core.src.layers.preprocessing.normalization import Normalization
from keras_core.src.layers.preprocessing.random_brightness import RandomBrightness
from keras_core.src.layers.preprocessing.random_contrast import RandomContrast
from keras_core.src.layers.preprocessing.random_crop import RandomCrop
from keras_core.src.layers.preprocessing.random_flip import RandomFlip
from keras_core.src.layers.preprocessing.random_rotation import RandomRotation
from keras_core.src.layers.preprocessing.random_translation import RandomTranslation
from keras_core.src.layers.preprocessing.random_zoom import RandomZoom
from keras_core.src.layers.preprocessing.rescaling import Rescaling
from keras_core.src.layers.preprocessing.resizing import Resizing
from keras_core.src.layers.preprocessing.string_lookup import StringLookup
from keras_core.src.layers.preprocessing.text_vectorization import TextVectorization
from keras_core.src.layers.regularization.activity_regularization import (
    ActivityRegularization,
)
from keras_core.src.layers.regularization.dropout import Dropout
from keras_core.src.layers.regularization.gaussian_dropout import GaussianDropout
from keras_core.src.layers.regularization.gaussian_noise import GaussianNoise
from keras_core.src.layers.regularization.spatial_dropout import SpatialDropout1D
from keras_core.src.layers.regularization.spatial_dropout import SpatialDropout2D
from keras_core.src.layers.regularization.spatial_dropout import SpatialDropout3D
from keras_core.src.layers.reshaping.cropping1d import Cropping1D
from keras_core.src.layers.reshaping.cropping2d import Cropping2D
from keras_core.src.layers.reshaping.cropping3d import Cropping3D
from keras_core.src.layers.reshaping.flatten import Flatten
from keras_core.src.layers.reshaping.permute import Permute
from keras_core.src.layers.reshaping.repeat_vector import RepeatVector
from keras_core.src.layers.reshaping.reshape import Reshape
from keras_core.src.layers.reshaping.up_sampling1d import UpSampling1D
from keras_core.src.layers.reshaping.up_sampling2d import UpSampling2D
from keras_core.src.layers.reshaping.up_sampling3d import UpSampling3D
from keras_core.src.layers.reshaping.zero_padding1d import ZeroPadding1D
from keras_core.src.layers.reshaping.zero_padding2d import ZeroPadding2D
from keras_core.src.layers.reshaping.zero_padding3d import ZeroPadding3D
from keras_core.src.layers.rnn.bidirectional import Bidirectional
from keras_core.src.layers.rnn.conv_lstm1d import ConvLSTM1D
from keras_core.src.layers.rnn.conv_lstm2d import ConvLSTM2D
from keras_core.src.layers.rnn.conv_lstm3d import ConvLSTM3D
from keras_core.src.layers.rnn.gru import GRU
from keras_core.src.layers.rnn.lstm import LSTM
from keras_core.src.layers.rnn.rnn import RNN
from keras_core.src.layers.rnn.simple_rnn import SimpleRNN
from keras_core.src.layers.rnn.stacked_rnn_cells import StackedRNNCells
from keras_core.src.layers.rnn.time_distributed import TimeDistributed
from keras_core.src.saving import serialization_lib


@keras_core_export("keras_core.layers.serialize")
def serialize(layer):
    """Returns the layer configuration as a Python dict.

    Args:
        layer: A `keras.layers.Layer` instance to serialize.

    Returns:
        Python dict which contains the configuration of the layer.
    """
    return serialization_lib.serialize_keras_object(layer)


@keras_core_export("keras_core.layers.deserialize")
def deserialize(config, custom_objects=None):
    """Returns a Keras layer object via its configuration.

    Args:
        config: A python dict containing a serialized layer configuration.
        custom_objects: Optional dictionary mapping names (strings) to custom
            objects (classes and functions) to be considered during
            deserialization.

    Returns:
        A Keras layer instance.
    """
    obj = serialization_lib.deserialize_keras_object(
        config,
        custom_objects=custom_objects,
    )
    if not isinstance(obj, Layer):
        raise ValueError(
            "`keras.layers.deserialize` was passed a `config` object that is "
            f"not a `keras.layers.Layer`. Received: {config}"
        )
    return obj




from keras_core.src import backend
from keras_core.src.api_export import keras_core_export
from keras_core.src.layers.layer import Layer
from keras_core.src.utils import backend_utils
from keras_core.src.utils.module_utils import tensorflow as tf


@keras_core_export("keras_core.layers.HashedCrossing")
class HashedCrossing(Layer):
    """A preprocessing layer which crosses features using the "hashing trick".

    This layer performs crosses of categorical features using the "hashing
    trick". Conceptually, the transformation can be thought of as:
    `hash(concatenate(features)) % num_bins.

    This layer currently only performs crosses of scalar inputs and batches of
    scalar inputs. Valid input shapes are `(batch_size, 1)`, `(batch_size,)` and
    `()`.

    **Note:** This layer wraps `tf.keras.layers.HashedCrossing`. It cannot
    be used as part of the compiled computation graph of a model with
    any backend other than TensorFlow.
    It can however be used with any backend when running eagerly.
    It can also always be used as part of an input preprocessing pipeline
    with any backend (outside the model itself), which is how we recommend
    to use this layer.

    **Note:** This layer is safe to use inside a `tf.data` pipeline
    (independently of which backend you're using).

    Args:
        num_bins: Number of hash bins.
        output_mode: Specification for the output of the layer. Values can be
            `"int"`, or `"one_hot"` configuring the layer as follows:
            - `"int"`: Return the integer bin indices directly.
            - `"one_hot"`: Encodes each individual element in the input into an
                array the same size as `num_bins`, containing a 1 at the input's
                bin index. Defaults to `"int"`.
        sparse: Boolean. Only applicable to `"one_hot"` mode and only valid
            when using the TensorFlow backend. If `True`, returns
            a `SparseTensor` instead of a dense `Tensor`. Defaults to `False`.
        **kwargs: Keyword arguments to construct a layer.

    Examples:

    **Crossing two scalar features.**

    >>> layer = keras_core.layers.HashedCrossing(
    ...     num_bins=5)
    >>> feat1 = np.array(['A', 'B', 'A', 'B', 'A'])
    >>> feat2 = np.array([101, 101, 101, 102, 102])
    >>> layer((feat1, feat2))
    array([1, 4, 1, 1, 3])

    **Crossing and one-hotting two scalar features.**

    >>> layer = keras_core.layers.HashedCrossing(
    ...     num_bins=5, output_mode='one_hot')
    >>> feat1 = np.array(['A', 'B', 'A', 'B', 'A'])
    >>> feat2 = np.array([101, 101, 101, 102, 102])
    >>> layer((feat1, feat2))
    array([[0., 1., 0., 0., 0.],
            [0., 0., 0., 0., 1.],
            [0., 1., 0., 0., 0.],
            [0., 1., 0., 0., 0.],
            [0., 0., 0., 1., 0.]], dtype=float32)
    """

    def __init__(
        self,
        num_bins,
        output_mode="int",
        sparse=False,
        name=None,
        dtype=None,
        **kwargs,
    ):
        if not tf.available:
            raise ImportError(
                "Layer HashedCrossing requires TensorFlow. "
                "Install it via `pip install tensorflow`."
            )

        if output_mode == "int" and dtype is None:
            dtype = "int64"
        super().__init__(name=name, dtype=dtype)
        if sparse and backend.backend() != "tensorflow":
            raise ValueError(
                "`sparse` can only be set to True with the "
                "TensorFlow backend."
            )
        self.layer = tf.keras.layers.HashedCrossing(
            num_bins=num_bins,
            output_mode=output_mode,
            sparse=sparse,
            name=name,
            dtype=dtype,
            **kwargs,
        )
        self.num_bins = num_bins
        self.output_mode = output_mode
        self.sparse = sparse
        self._allow_non_tensor_positional_args = True
        self._convert_input_args = False
        self.supports_jit = False

    def compute_output_shape(self, input_shape):
        if (
            not len(input_shape) == 2
            or not isinstance(input_shape[0], tuple)
            or not isinstance(input_shape[1], tuple)
        ):
            raise ValueError(
                "Expected as input a list/tuple of 2 tensors. "
                f"Received input_shape={input_shape}"
            )
        if input_shape[0][-1] != input_shape[1][-1]:
            raise ValueError(
                "Expected the two input tensors to have identical shapes. "
                f"Received input_shape={input_shape}"
            )

        if not input_shape:
            if self.output_mode == "int":
                return ()
            return (self.num_bins,)
        if self.output_mode == "int":
            return input_shape[0]

        if self.output_mode == "one_hot" and input_shape[0][-1] != 1:
            output_shape = tuple(input_shape[0]) + (self.num_bins,)
        output_shape = tuple(input_shape[0])[:-1] + (self.num_bins,)
        return output_shape

    def call(self, inputs):
        outputs = self.layer.call(inputs)
        if (
            backend.backend() != "tensorflow"
            and not backend_utils.in_tf_graph()
        ):
            outputs = backend.convert_to_tensor(outputs)
        return outputs

    def get_config(self):
        return {
            "num_bins": self.num_bins,
            "output_mode": self.output_mode,
            "sparse": self.sparse,
            "name": self.name,
            "dtype": self.dtype,
        }




from keras_core.src import backend
from keras_core.src import layers
from keras_core.src import ops
from keras_core.src.api_export import keras_core_export


@keras_core_export("keras_core.layers.GaussianNoise")
class GaussianNoise(layers.Layer):
    """Apply additive zero-centered Gaussian noise.

    This is useful to mitigate overfitting
    (you could see it as a form of random data augmentation).
    Gaussian Noise (GS) is a natural choice as corruption process
    for real valued inputs.

    As it is a regularization layer, it is only active at training time.

    Args:
        stddev: Float, standard deviation of the noise distribution.
        seed: Integer, optional random seed to enable deterministic behavior.

    Call arguments:
        inputs: Input tensor (of any rank).
        training: Python boolean indicating whether the layer should behave in
            training mode (adding noise) or in inference mode (doing nothing).
    """

    def __init__(self, stddev, seed=None, **kwargs):
        super().__init__(**kwargs)
        if not 0 <= stddev <= 1:
            raise ValueError(
                f"Invalid value received for argument "
                "`stddev`. Expected a float value between 0 and 1. "
                f"Received: stddev={stddev}"
            )
        self.stddev = stddev
        self.seed = seed
        self.seed_generator = backend.random.SeedGenerator(seed)
        self.supports_masking = True

    def call(self, inputs, training=False):
        if training and self.stddev > 0:
            return inputs + backend.random.normal(
                shape=ops.shape(inputs),
                mean=0.0,
                stddev=self.stddev,
                seed=self.seed_generator,
            )
        return inputs

    def compute_output_shape(self, input_shape):
        return input_shape

    def get_config(self):
        base_config = super().get_config()
        config = {
            "stddev": self.stddev,
            "seed": self.seed,
        }
        return {**base_config, **config}




import numpy as np

from keras_core.src import backend
from keras_core.src.api_export import keras_core_export
from keras_core.src.layers.layer import Layer
from keras_core.src.utils import backend_utils
from keras_core.src.utils.module_utils import tensorflow as tf


@keras_core_export("keras_core.layers.Discretization")
class Discretization(Layer):
    """A preprocessing layer which buckets continuous features by ranges.

    This layer will place each element of its input data into one of several
    contiguous ranges and output an integer index indicating which range each
    element was placed in.

    **Note:** This layer wraps `tf.keras.layers.Discretization`. It cannot
    be used as part of the compiled computation graph of a model with
    any backend other than TensorFlow.
    It can however be used with any backend when running eagerly.
    It can also always be used as part of an input preprocessing pipeline
    with any backend (outside the model itself), which is how we recommend
    to use this layer.

    **Note:** This layer is safe to use inside a `tf.data` pipeline
    (independently of which backend you're using).

    Input shape:
        Any array of dimension 2 or higher.

    Output shape:
        Same as input shape.

    Arguments:
        bin_boundaries: A list of bin boundaries.
            The leftmost and rightmost bins
            will always extend to `-inf` and `inf`,
            so `bin_boundaries=[0., 1., 2.]`
            generates bins `(-inf, 0.)`, `[0., 1.)`, `[1., 2.)`,
            and `[2., +inf)`.
            If this option is set, `adapt()` should not be called.
        num_bins: The integer number of bins to compute.
            If this option is set,
            `adapt()` should be called to learn the bin boundaries.
        epsilon: Error tolerance, typically a small fraction
            close to zero (e.g. 0.01). Higher values of epsilon increase
            the quantile approximation, and hence result in more
            unequal buckets, but could improve performance
            and resource consumption.
        output_mode: Specification for the output of the layer.
            Values can be `"int"`, `"one_hot"`, `"multi_hot"`, or
            `"count"` configuring the layer as follows:
            - `"int"`: Return the discretized bin indices directly.
            - `"one_hot"`: Encodes each individual element in the
                input into an array the same size as `num_bins`,
                containing a 1 at the input's bin
                index. If the last dimension is size 1, will encode on that
                dimension.  If the last dimension is not size 1,
                will append a new dimension for the encoded output.
            - `"multi_hot"`: Encodes each sample in the input into a
                single array the same size as `num_bins`,
                containing a 1 for each bin index
                index present in the sample.
                Treats the last dimension as the sample
                dimension, if input shape is `(..., sample_length)`,
                output shape will be `(..., num_tokens)`.
            - `"count"`: As `"multi_hot"`, but the int array contains
                a count of the number of times the bin index appeared
                in the sample.
            Defaults to `"int"`.
        sparse: Boolean. Only applicable to `"one_hot"`, `"multi_hot"`,
            and `"count"` output modes. Only supported with TensorFlow
            backend. If `True`, returns a `SparseTensor` instead of
            a dense `Tensor`. Defaults to `False`.

    Examples:

    Bucketize float values based on provided buckets.
    >>> input = np.array([[-1.5, 1.0, 3.4, .5], [0.0, 3.0, 1.3, 0.0]])
    >>> layer = Discretization(bin_boundaries=[0., 1., 2.])
    >>> layer(input)
    array([[0, 2, 3, 1],
           [1, 3, 2, 1]])

    Bucketize float values based on a number of buckets to compute.
    >>> input = np.array([[-1.5, 1.0, 3.4, .5], [0.0, 3.0, 1.3, 0.0]])
    >>> layer = Discretization(num_bins=4, epsilon=0.01)
    >>> layer.adapt(input)
    >>> layer(input)
    array([[0, 2, 3, 2],
           [1, 3, 3, 1]])
    """

    def __init__(
        self,
        bin_boundaries=None,
        num_bins=None,
        epsilon=0.01,
        output_mode="int",
        sparse=False,
        name=None,
        dtype=None,
        **kwargs,
    ):
        if not tf.available:
            raise ImportError(
                "Layer Discretization requires TensorFlow. "
                "Install it via `pip install tensorflow`."
            )

        super().__init__(name=name, dtype=dtype)
        if sparse and backend.backend() != "tensorflow":
            raise ValueError(
                "`sparse` can only be set to True with the "
                "TensorFlow backend."
            )
        self.layer = tf.keras.layers.Discretization(
            bin_boundaries=bin_boundaries,
            num_bins=num_bins,
            epsilon=epsilon,
            output_mode=output_mode,
            sparse=sparse,
            name=name,
            dtype=dtype,
            **kwargs,
        )
        self.bin_boundaries = bin_boundaries
        if self.bin_boundaries:
            self.built = True
        self._convert_input_args = False
        self._allow_non_tensor_positional_args = True

        self.num_bins = num_bins
        self.epsilon = epsilon
        self.output_mode = output_mode
        self.sparse = sparse
        self.supports_jit = False

    def build(self, input_shape):
        self.layer.build(input_shape)
        self.built = True

    # We override this method solely to generate a docstring.
    def adapt(self, data, batch_size=None, steps=None):
        """Computes bin boundaries from quantiles in a input dataset.

        Calling `adapt()` on a `Discretization` layer is an alternative to
        passing in a `bin_boundaries` argument during construction. A
        `Discretization` layer should always be either adapted over a dataset or
        passed `bin_boundaries`.

        During `adapt()`, the layer will estimate the quantile boundaries of the
        input dataset. The number of quantiles can be controlled via the
        `num_bins` argument, and the error tolerance for quantile boundaries can
        be controlled via the `epsilon` argument.

        Arguments:
            data: The data to train on. It can be passed either as a
                batched `tf.data.Dataset`,
                or as a NumPy array.
            batch_size: Integer or `None`.
                Number of samples per state update.
                If unspecified, `batch_size` will default to 32.
                Do not specify the `batch_size` if your data is in the
                form of a `tf.data.Dataset`
                (it is expected to be already batched).
            steps: Integer or `None`.
                Total number of steps (batches of samples)
                When training with input tensors such as
                the default `None` is equal to
                the number of samples in your dataset divided by
                the batch size, or 1 if that cannot be determined.
                If `data` is a `tf.data.Dataset`, and `steps` is `None`,
                `adapt()` will run until the input dataset is exhausted.
                When passing an infinitely
                repeating dataset, you must specify the `steps` argument. This
                argument is not supported with array inputs or list inputs.
        """
        self.layer.adapt(data, batch_size=batch_size, steps=steps)

    def update_state(self, data):
        self.layer.update_state(data)

    def finalize_state(self):
        self.layer.finalize_state()

    def reset_state(self):
        self.layer.reset_state()

    def compute_output_spec(self, inputs):
        return backend.KerasTensor(shape=inputs.shape, dtype="int32")

    def __call__(self, inputs):
        if not isinstance(inputs, (tf.Tensor, np.ndarray, backend.KerasTensor)):
            inputs = tf.convert_to_tensor(backend.convert_to_numpy(inputs))
        if not self.built:
            self.build(inputs.shape)
        return super().__call__(inputs)

    def call(self, inputs):
        outputs = self.layer.call(inputs)
        if (
            backend.backend() != "tensorflow"
            and not backend_utils.in_tf_graph()
        ):
            outputs = backend.convert_to_tensor(outputs)
        return outputs

    def get_config(self):
        return {
            "bin_boundaries": self.bin_boundaries,
            "num_bins": self.num_bins,
            "epsilon": self.epsilon,
            "output_mode": self.output_mode,
            "sparse": self.sparse,
            "name": self.name,
            "dtype": self.dtype,
        }




from keras_core.src import ops
from keras_core.src.api_export import keras_core_export
from keras_core.src.layers.input_spec import InputSpec
from keras_core.src.layers.layer import Layer
from keras_core.src.utils import argument_validation


@keras_core_export("keras_core.layers.ZeroPadding1D")
class ZeroPadding1D(Layer):
    """Zero-padding layer for 1D input (e.g. temporal sequence).

    Examples:

    >>> input_shape = (2, 2, 3)
    >>> x = np.arange(np.prod(input_shape)).reshape(input_shape)
    >>> x
    [[[ 0  1  2]
      [ 3  4  5]]
     [[ 6  7  8]
      [ 9 10 11]]]
    >>> y = keras_core.layers.ZeroPadding1D(padding=2)(x)
    >>> y
    [[[ 0  0  0]
      [ 0  0  0]
      [ 0  1  2]
      [ 3  4  5]
      [ 0  0  0]
      [ 0  0  0]]
     [[ 0  0  0]
      [ 0  0  0]
      [ 6  7  8]
      [ 9 10 11]
      [ 0  0  0]
      [ 0  0  0]]]

    Args:
        padding: Int, or tuple of int (length 2), or dictionary.
            - If int: how many zeros to add at the beginning and end of
              the padding dimension (axis 1).
            - If tuple of 2 ints: how many zeros to add at the beginning and the
              end of the padding dimension (`(left_pad, right_pad)`).

    Input shape:
        3D tensor with shape `(batch_size, axis_to_pad, features)`

    Output shape:
        3D tensor with shape `(batch_size, padded_axis, features)`
    """

    def __init__(self, padding=1, **kwargs):
        super().__init__(**kwargs)
        self.padding = argument_validation.standardize_tuple(
            padding, 2, "padding", allow_zero=True
        )
        self.input_spec = InputSpec(ndim=3)

    def compute_output_shape(self, input_shape):
        output_shape = list(input_shape)
        if output_shape[1] is not None:
            output_shape[1] += self.padding[0] + self.padding[1]
        return tuple(output_shape)

    def call(self, inputs):
        all_dims_padding = ((0, 0), self.padding, (0, 0))
        return ops.pad(inputs, all_dims_padding)

    def get_config(self):
        config = {"padding": self.padding}
        base_config = super().get_config()
        return {**base_config, **config}




import torch
import torch.nn.functional as tnn

from keras_core.src.backend.torch.core import convert_to_tensor

RESIZE_INTERPOLATIONS = {}  # populated after torchvision import

UNSUPPORTED_INTERPOLATIONS = (
    "lanczos3",
    "lanczos5",
)


def resize(
    image,
    size,
    interpolation="bilinear",
    antialias=False,
    data_format="channels_last",
):
    try:
        import torchvision
        from torchvision.transforms import InterpolationMode as im

        RESIZE_INTERPOLATIONS.update(
            {
                "bilinear": im.BILINEAR,
                "nearest": im.NEAREST_EXACT,
                "bicubic": im.BICUBIC,
            }
        )
    except:
        raise ImportError(
            "The torchvision package is necessary to use `resize` with the "
            "torch backend. Please install torchvision."
        )
    if interpolation in UNSUPPORTED_INTERPOLATIONS:
        raise ValueError(
            "Resizing with Lanczos interpolation is "
            "not supported by the PyTorch backend. "
            f"Received: interpolation={interpolation}."
        )
    if interpolation not in RESIZE_INTERPOLATIONS:
        raise ValueError(
            "Invalid value for argument `interpolation`. Expected of one "
            f"{RESIZE_INTERPOLATIONS}. Received: interpolation={interpolation}"
        )
    if not len(size) == 2:
        raise ValueError(
            "Argument `size` must be a tuple of two elements "
            f"(height, width). Received: size={size}"
        )
    size = tuple(size)
    image = convert_to_tensor(image)
    if data_format == "channels_last":
        if image.ndim == 4:
            image = image.permute((0, 3, 1, 2))
        elif image.ndim == 3:
            image = image.permute((2, 0, 1))
        else:
            raise ValueError(
                "Invalid input rank: expected rank 3 (single image) "
                "or rank 4 (batch of images). Received input with shape: "
                f"image.shape={image.shape}"
            )

    resized = torchvision.transforms.functional.resize(
        img=image,
        size=size,
        interpolation=RESIZE_INTERPOLATIONS[interpolation],
        antialias=antialias,
    )
    if data_format == "channels_last":
        if len(image.shape) == 4:
            resized = resized.permute((0, 2, 3, 1))
        elif len(image.shape) == 3:
            resized = resized.permute((1, 2, 0))
    return resized


AFFINE_TRANSFORM_INTERPOLATIONS = (
    "nearest",
    "bilinear",
)
AFFINE_TRANSFORM_FILL_MODES = {
    "constant": "zeros",
    "nearest": "border",
    # "wrap",  not supported by torch
    "mirror": "reflection",  # torch's reflection is mirror in other backends
    "reflect": "reflection",  # if fill_mode==reflect, redirect to mirror
}


def _apply_grid_transform(
    img,
    grid,
    interpolation="bilinear",
    fill_mode="zeros",
    fill_value=None,
):
    """
    Modified from https://github.com/pytorch/vision/blob/main/torchvision/transforms/v2/functional/_geometry.py
    """  # noqa: E501

    # We are using context knowledge that grid should have float dtype
    fp = img.dtype == grid.dtype
    float_img = img if fp else img.to(grid.dtype)

    shape = float_img.shape
    # Append a dummy mask for customized fill colors, should be faster than
    # grid_sample() twice
    if fill_value is not None:
        mask = torch.ones(
            (shape[0], 1, shape[2], shape[3]),
            dtype=float_img.dtype,
            device=float_img.device,
        )
        float_img = torch.cat((float_img, mask), dim=1)

    float_img = tnn.grid_sample(
        float_img,
        grid,
        mode=interpolation,
        padding_mode=fill_mode,
        align_corners=True,
    )
    # Fill with required color
    if fill_value is not None:
        float_img, mask = torch.tensor_split(float_img, indices=(-1,), dim=-3)
        mask = mask.expand_as(float_img)
        fill_list = (
            fill_value
            if isinstance(fill_value, (tuple, list))
            else [float(fill_value)]
        )
        fill_img = torch.tensor(
            fill_list, dtype=float_img.dtype, device=float_img.device
        ).view(1, -1, 1, 1)
        if interpolation == "nearest":
            bool_mask = mask < 0.5
            float_img[bool_mask] = fill_img.expand_as(float_img)[bool_mask]
        else:  # 'bilinear'
            # The following is mathematically equivalent to:
            # img * mask + (1.0 - mask) * fill =
            # img * mask - fill * mask + fill =
            # mask * (img - fill) + fill
            float_img = float_img.sub_(fill_img).mul_(mask).add_(fill_img)

    img = float_img.round_().to(img.dtype) if not fp else float_img
    return img


def affine_transform(
    image,
    transform,
    interpolation="bilinear",
    fill_mode="constant",
    fill_value=0,
    data_format="channels_last",
):
    if interpolation not in AFFINE_TRANSFORM_INTERPOLATIONS:
        raise ValueError(
            "Invalid value for argument `interpolation`. Expected of one "
            f"{AFFINE_TRANSFORM_INTERPOLATIONS}. Received: "
            f"interpolation={interpolation}"
        )
    if fill_mode not in AFFINE_TRANSFORM_FILL_MODES.keys():
        raise ValueError(
            "Invalid value for argument `fill_mode`. Expected of one "
            f"{set(AFFINE_TRANSFORM_FILL_MODES.keys())}. "
            f"Received: fill_mode={fill_mode}"
        )

    image = convert_to_tensor(image)
    transform = convert_to_tensor(transform)

    if image.ndim not in (3, 4):
        raise ValueError(
            "Invalid image rank: expected rank 3 (single image) "
            "or rank 4 (batch of images). Received input with shape: "
            f"image.shape={image.shape}"
        )
    if transform.ndim not in (1, 2):
        raise ValueError(
            "Invalid transform rank: expected rank 1 (single transform) "
            "or rank 2 (batch of transforms). Received input with shape: "
            f"transform.shape={transform.shape}"
        )

    # the default fill_value of tnn.grid_sample is "zeros"
    if fill_mode != "constant" or (fill_mode == "constant" and fill_value == 0):
        fill_value = None

    # unbatched case
    need_squeeze = False
    if image.ndim == 3:
        image = image.unsqueeze(dim=0)
        need_squeeze = True
    if transform.ndim == 1:
        transform = transform.unsqueeze(dim=0)

    if data_format == "channels_last":
        image = image.permute((0, 3, 1, 2))

    batch_size = image.shape[0]
    h, w, c = image.shape[-2], image.shape[-1], image.shape[-3]

    # get indices
    shape = [h, w, c]  # (H, W, C)
    meshgrid = torch.meshgrid(
        *[torch.arange(size) for size in shape], indexing="ij"
    )
    indices = torch.concatenate(
        [torch.unsqueeze(x, dim=-1) for x in meshgrid], dim=-1
    )
    indices = torch.tile(indices, (batch_size, 1, 1, 1, 1))
    indices = indices.to(transform)

    # swap the values
    a0 = transform[:, 0].clone()
    a2 = transform[:, 2].clone()
    b1 = transform[:, 4].clone()
    b2 = transform[:, 5].clone()
    transform[:, 0] = b1
    transform[:, 2] = b2
    transform[:, 4] = a0
    transform[:, 5] = a2

    # deal with transform
    transform = torch.nn.functional.pad(
        transform, pad=[0, 1, 0, 0], mode="constant", value=1
    )
    transform = torch.reshape(transform, (batch_size, 3, 3))
    offset = transform[:, 0:2, 2].clone()
    offset = torch.nn.functional.pad(offset, pad=[0, 1, 0, 0])
    transform[:, 0:2, 2] = 0

    # transform the indices
    coordinates = torch.einsum("Bhwij, Bjk -> Bhwik", indices, transform)
    coordinates = torch.moveaxis(coordinates, source=-1, destination=1)
    coordinates += torch.reshape(a=offset, shape=(*offset.shape, 1, 1, 1))
    coordinates = coordinates[:, 0:2, ..., 0]
    coordinates = coordinates.permute((0, 2, 3, 1))

    # normalize coordinates
    coordinates[:, :, :, 1] = coordinates[:, :, :, 1] / (w - 1) * 2.0 - 1.0
    coordinates[:, :, :, 0] = coordinates[:, :, :, 0] / (h - 1) * 2.0 - 1.0
    grid = torch.stack(
        [coordinates[:, :, :, 1], coordinates[:, :, :, 0]], dim=-1
    )

    affined = _apply_grid_transform(
        image,
        grid,
        interpolation=interpolation,
        # if fill_mode==reflect, redirect to mirror
        fill_mode=AFFINE_TRANSFORM_FILL_MODES[fill_mode],
        fill_value=fill_value,
    )

    if data_format == "channels_last":
        affined = affined.permute((0, 2, 3, 1))
    if need_squeeze:
        affined = affined.squeeze(dim=0)
    return affined




from keras_core.src import backend
from keras_core.src.api_export import keras_core_export

if backend.backend() == "tensorflow":
    BackendVariable = backend.tensorflow.core.Variable
elif backend.backend() == "jax":
    BackendVariable = backend.jax.core.Variable
elif backend.backend() == "torch":
    BackendVariable = backend.torch.core.Variable
elif backend.backend() == "numpy":
    from keras_core.src.backend.numpy.core import Variable as NumpyVariable

    BackendVariable = NumpyVariable
else:
    raise RuntimeError("Invalid backend.")


@keras_core_export("keras_core.backend.Variable")
class Variable(BackendVariable):
    pass

